/*
 * SupportWizard 2
 * 
 * Copyright (C) 2001 ISC corp. All Rights Reserved.
 * 
 * $Id: SWFunctionalitiesOrder.java,v 1.13 2003/07/22 23:05:30 alexi Exp $
 * Created by Konstantin Sobolev (kos@supportwizard.com)
 * Date: Oct 3, 2001
 * Time: 6:14:08 PM
 * 
 * Last modification $Date: 2003/07/22 23:05:30 $
 */
package com.supportwizard.functionalities;

import com.supportwizard.functionalities.history.SWHistoryFunctionalityMeta;
import com.supportwizard.functionalities.workflow.SWWorkflowFunctionalityMeta;
import com.supportwizard.functionalities.choice.SWChoiceFunctionalityMeta;
import com.supportwizard.functionalities.owner.ejb.OwnerFunctionalityConst;
import com.supportwizard.functionalities.calculation.SWCalculationFunctionalityMeta;
import java.rmi.RemoteException;
import java.util.*;

/**
 * This class determines functionalities invocation order. This order is hard-coded and one
 * must adjust <code>FUNC_PRIORITIES</code> set to change it. It is only used by
 * <code>SWFunctionalitiesEventBroadcasterBean.getEventTargets(SWTable)</code> method.
 * 
 * @see com.supportwizard.functionalities.ejb.SWFunctionalityEventBroadcasterBean
 * 
 * @author Konstantin Sobolev (kos@supportwizard.com)
 * @version $Revision: 1.13 $
 */
public final class SWFunctionalitiesOrder {

  /**
   * Functionalities priorities set. Contains functionality-name -- callback -- priority tuples.
   * Functionality name is a <code>String</code> that can be obtained using
   * <code>func.getFunctionalityName()</code> call. Callback is a method name as
   * defined in <code>SWFunctionalityLogic</code> or <code>null</code> which
   * means all methods. The same functionality may appear several times
   * with varisou method specifications. Priority is an integer.<p/>
   * The bigger priority is, the earlier functionality will be called. Negative values allowed.
   */
  public static final Set FUNC_PRIORITIES = new HashSet();
  static {

    // fill priorities;
    FUNC_PRIORITIES.add(new FuncPriorityTuple(SWHistoryFunctionalityMeta.FUNC_NAME, -1));  // must be the last one! (smallest)

    // workflow functionality should run after rules2 functionality:
    FUNC_PRIORITIES.add(new FuncPriorityTuple(SWWorkflowFunctionalityMeta.FUNC_NAME, 1));
    FUNC_PRIORITIES.add(new FuncPriorityTuple(SWCalculationFunctionalityMeta.FUNC_NAME, 8));

    // choice functionality should run after owner f-ty
    FUNC_PRIORITIES.add(new FuncPriorityTuple(SWChoiceFunctionalityMeta.FUNC_NAME, 9));
    FUNC_PRIORITIES.add(new FuncPriorityTuple(OwnerFunctionalityConst.FUNC_NAME, 10));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Contact", 11));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Subtype", 12));

    // for tests
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Dummy1", 1));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Dummy1", "roDataReading", 1));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Dummy2", "roDataReading", -1));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Dummy1", "rwDataReading", -1));
    FUNC_PRIORITIES.add(new FuncPriorityTuple("Dummy2", "rwDataReading", 1));
  } 

  /**
   * Sorts functionalities according to <code>FUNC_PRIORITIES</code> map.
   * Functionalities will be sorted in <em>descending</em> order, so the first
   * functionality will be functionality with max priority. If some functionality is not included
   * into <code>FUNC_PRIORITIES</code> map then it will be assigned priority of zero.
   * 
   * @param funcs Collection of functionalities to be sorted.
   * @param callback name of a callback methond that is going to be called.
   * @return Sorted list of functionalities.
   */
  public static List sortFunctionalities(Collection funcs, String callback) throws RemoteException {
    HashMap namesToFuncs = buildNamesToFuncs(funcs);
    List    matchingTuples = getMatchingTuples(namesToFuncs.keySet(), callback);

    Collections.sort(matchingTuples);

    return convertTuplesToFunctionalities(matchingTuples, namesToFuncs);
  } 

  private static List getMatchingTuples(Collection funcNames, String callback) {
    LinkedList res = new LinkedList();

    for (Iterator it = funcNames.iterator(); it.hasNext(); ) {
      String funcName = (String) it.next();

      res.add(findMatchingTuple(funcName, callback));
    } 

    return res;
  } 

  // finds best mathing tuple from FUNC_PRIORITIES
  private static FuncPriorityTuple findMatchingTuple(String funcName, String callback) {
    FuncPriorityTuple res = new FuncPriorityTuple(funcName, 0);

    for (Iterator it = FUNC_PRIORITIES.iterator(); it.hasNext(); ) {
      FuncPriorityTuple tuple = (FuncPriorityTuple) it.next();

      if (!tuple.getFuncName().equals(funcName)) {
        continue;      // wrong functionality name
      } 

      if (tuple.getCallback() != null &&!callback.equals(tuple.getCallback())) {
        continue;      // wrong callback name
      } 

      if (callback.equals(tuple.getCallback())) {
        return tuple;  // exact match
      } 

      res = tuple;     // partial match
    } 

    return res;
  } 

  private static List convertTuplesToFunctionalities(List filteredTuples, HashMap namesToFuncs) {
    ArrayList res = new ArrayList(filteredTuples.size());

    for (Iterator it = filteredTuples.iterator(); it.hasNext(); ) {
      FuncPriorityTuple tuple = (FuncPriorityTuple) it.next();
      final List        fl = (List) namesToFuncs.get(tuple.getFuncName());

      res.addAll(fl);
    } 

    return res;
  } 

  // builds map 'funcName' -> list of functionalities with this name
  private static HashMap buildNamesToFuncs(Collection funcs) throws RemoteException {
    HashMap res = new HashMap(funcs.size());

    for (Iterator it = funcs.iterator(); it.hasNext(); ) {
      SWFunctionality functionality = (SWFunctionality) it.next();
      final String    funcName = functionality.getFunctionalityName();
      List            fl = (List) res.get(funcName);

      if (fl == null) {
        fl = new ArrayList(3);
      } 

      fl.add(functionality);
      res.put(funcName, fl);
    } 

    return res;
  } 

  /**
   * Functionality--callback--priority tuple.
   * 
   * @see java.lang.Comparable
   */
  static public class FuncPriorityTuple implements Comparable {
    private String funcName;
    private String callback;
    private int    priority;

    /**
     * Creates new instance that will match every method of specified functionality.
     * 
     * @param funcName Functionality name.
     * @param priority Priority.
     */
    public FuncPriorityTuple(String funcName, int priority) {
      this.funcName = funcName;
      this.callback = null;
      this.priority = priority;
    }

    /**
     * Creates new instance that will match only one method of specified functionality.
     * 
     * @param funcName Functionality name.
     * @param callback Callback method name. If <code>null</code>, then all methods will be matched.
     * @param priority Priority.
     */
    public FuncPriorityTuple(String funcName, String callback, int priority) {
      this.callback = callback;
      this.funcName = funcName;
      this.priority = priority;
    }

    /**
     * Gets functionality name.
     * 
     * @return functionality name.
     */
    String getFuncName() {
      return funcName;
    } 

    /**
     * Gets callback method name.
     * 
     * @return callback method name.
     */
    String getCallback() {
      return callback;
    } 

    /**
     * Compares this object with the specified object for order.  Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     */
    public int compareTo(Object o) {
      FuncPriorityTuple p = (FuncPriorityTuple) o;

      if (priority > p.priority) {
        return -1;  // actually we need reverse order i.e. 99,98,97,...,0,0,0,0
      } 

      if (priority < p.priority) {
        return 1;
      } 

      return 0;
    } 

  }

}

