/*
 * SupportWizard 2
 * 
 * Copyright (C) 2001 ISC corp. All Rights Reserved.
 * 
 * $Id: SWFunctionalityBase.java,v 1.73 2003/09/03 18:47:58 kos Exp $
 * 
 * Created by Konstantin Sobolev.
 * Date: Jul 12, 2001
 * Time: 11:54:20 AM
 */
package com.supportwizard.functionalities;

import com.supportwizard.cache.SimpleCache;
import com.supportwizard.cache.SimpleCacheFactory;
import com.supportwizard.db.*;
import com.supportwizard.dictionary.SWTable;
import com.supportwizard.dictionary.SWTableHome;
import com.supportwizard.functionalities.interfaces.*;
import com.supportwizard.seance.Seance;
import com.supportwizard.utils.EJBException2;
import com.supportwizard.utils.HomesGetter;
import com.supportwizard.cachewrappers.SWTableRelatedCachesWrapper;
import org.apache.log4j.Logger;

import javax.ejb.*;
import javax.naming.NamingException;
import java.rmi.RemoteException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

/**
 * This is the desired base class for all functionalities implementations. It contains
 * all EJB-specific methods implementations. Of course if you want to write all BMP
 * logic yourself you may not inherit from this class, but in this case you're on your own!
 * 
 * The only method left to implement by real functionalitiy is
 * {@link #getFunctionalityCreationInfo() getFunctionalityCreationInfo()}. And of course, if you don't want to
 * create an empty functionality, you should override {@link #getFunctionalityLogic getFunctionalityLogic()}.
 * 
 * Note that this class provides some useful public and protected methods what can be used by functionalities implementations:
 * <ul>
 * <li>{@link #getFuncColsMappings() getFuncColsMappings()}</li>
 * <li>{@link #getFuncColsMapping(String) getFuncColsMapping(String)}</li>
 * <li>{@link #getParameters() getParameter()}</li>
 * <li>{@link #getParameter(String) getParameter(String)}</li>
 * <li>{@link #getFunctionalityID() getFunctionalityID()}</li>
 * <li>{@link #getFunctionalityMetadataID() getFunctionalityMetadataID()}</li>
 * <li>{@link #getFuncColsMappingHome() getFuncColsMappingHome()}</li>
 * <li>{@link #getFuncParameterHome() getFuncParameterHome()}</li>
 * </ul>
 * 
 * Functionalities are mapped to swfunctionalities table with following layout:
 * <pre>
 * +--------------+------------+------+-----+---------+----------------+---------------------------------+
 * | Field        | Type       | Null | Key | Default | Extra          | Privileges                      |
 * +--------------+------------+------+-----+---------+----------------+---------------------------------+
 * | SWFuncID     | bigint(20) |      | PRI |         | auto_increment | select,insert,update,references |
 * | SWFuncMetaID | bigint(20) | YES  |     |         |                | select,insert,update,references |
 * | SWTableID    | bigint(20) |      |     | 0       |                | select,insert,update,references |
 * | Enabled      | char(1)    |      |     |         |                | select,insert,update,references |
 * +--------------+------------+------+-----+---------+----------------+---------------------------------+
 * </pre>
 * 
 * @author Konstantin Sobolev
 * @version $Revision: 1.73 $
 * 
 * @ejb:ejb-ref ejb-name="SWFuncParameter"
 * @ejb:ejb-ref ejb-name="SWFuncColsMapping"
 * @ejb:ejb-ref ejb-name="SWFunctionalitiesFinder"
 * 
 * @ejb:finder signature="java.util.Collection findAll()"
 * @ejb:finder signature="java.util.Collection findByTable (com.supportwizard.dictionary.SWTable table)"
 * @ejb:finder signature="java.util.Collection findEnabledByTable (com.supportwizard.dictionary.SWTable table)"
 */
public abstract class SWFunctionalityBase {
	private static final String TABLE_NAME = "swfunctionalities";
	private transient Logger log = Logger.getLogger(SWFunctionalityBase.class);
	private SWFunctionalityCreationInfo functionalityCreationInfo;
	private HomesGetter homesGetter;
	private transient SimpleCache cache;
	private transient SWTableRelatedCachesWrapper tableRelatedCachesWrapper;

	/**
	 * <code>true</code> if functionality is enabled; <code>false</code> otherwise.
	 */
	protected boolean enabled;

	/**
	 * Functionality name.
	 */
	protected String name;

	/**
	 * Functionality description.
	 */
	protected String description;

	/**
	 * Functionality home class name.
	 */
	protected String homeClassName;

	/**
	 * Functionality home JNDI name.
	 */
	protected String homeJndiName;

	/**
	 * Functionality ID.
	 */
	protected Long funcID;

	/**
	 * Functionality metadata ID.
	 */
	protected Long funcMetaID;

	/**
	 * User table this functionality is attached to.
	 */
	protected SWTable table;

	/**
	 * Entity context.
	 */
	protected EntityContext entityContext;

	/**
	 * Seance.
	 */
	protected Seance seance;

	/**
	 * @link dependency
	 * @label implicitly implements
	 */

	/* #SWFunctionality lnkSWFunctionality; */

	/**
	 * Call this method to mark that functionality internal state must be synchrnomized
	 * with the database. The implementation is generated by ejbdoclet.
	 */
	protected abstract void makeDirty();

	/**
	 * Gets functionality creation info. Must be implemented by successors. To functionalities creators: information
	 * will be cached here, you don't need to cache it in functionality code.
	 * 
	 * @return Functionality creation info.
	 */
	protected abstract SWFunctionalityCreationInfo getFunctionalityCreationInfo();

	/**
	 * Gets seance.
	 * 
	 * @return seance.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 * @deprecated SWFunctionality is an entity bean, thus it shouldn't be Seanceable. When all
	 * functionalities will get rid of this seance, "<code>implements Seanceable</code>" will be eliminated.
	 */
	public Seance getSeance() {
		if (seance == null) {
			log.error("Seance is not set");
		}

		return seance;
	}

	/**
	 * Sets seance.
	 * 
	 * @param seance seance.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 * @deprecated SWFunctionality is an entity bean, thus it shouldn't be Seanceable. When all
	 * functionalities will get rid of this seance, "<code>implements Seanceable</code>" will be eliminated.
	 */
	public void setSeance(Seance seance) {
		if (seance == null) {
			log.warn("Skipping an attempt to set null seance");

			return;
		}

		this.seance = seance;
	}

	/**
	 * Gets functionality state.
	 * Functionalities may be temporarily disabled. This is useful when you need
	 * to modify some tables from functionality callback. Disabling functionality
	 * in such case will avoid internal recursion.
	 * 
	 * @return <code>true</code> if functionality is enabled.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Sets functionality state.
	 * Functionalities may be temporarily disabled. This is useful when you need
	 * to modify some tables from functionality callback. Disabling functionality
	 * in such case will avoid internal recursion.
	 * 
	 * @param enabled New functionality state.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Required"
	 */
	public void setEnabled(boolean enabled) {
		if (this.enabled == enabled)
			return;

		this.enabled = enabled;

		clearTableRelatedCaches();
		makeDirty();
	}

	/**
	 * Gets functionality name.
	 * 
	 * @return Name of this functionality.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public String getFunctionalityName() {
		return name;
	}

	/**
	 * Gets functionality description.
	 * 
	 * @return Description of this functionality.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public String getFunctionalityDescription() {
		return description;
	}

	/**
	 * Gets functionality ID.
	 * 
	 * @return Functionality ID.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Long getFunctionalityID() {
		return funcID;
	}

	/**
	 * Gets ID of SWTable this functionality is associated with.
	 * 
	 * @return SWTable ID
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Long getSWTableID() {
		try {
			return (Long) table.getPrimaryKey();
		} catch (RemoteException e) {
			log.error("Exception caught", e);

			throw new EJBException(e);
		}
	}

	/**
	 * Gets functionality metadata ID.
	 * 
	 * @return Functionality metadata ID.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Long getFunctionalityMetadataID() {
		return funcMetaID;
	}

	/**
	 * Gets functionality logic.
	 * 
	 * @return Functionality logic instance.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public SWFunctionalityLogic getFunctionalityLogic() {
		return new SWFunctionalityLogicBase(funcID, getSWTableID());
	}

	/**
	 * Gets functionality parameter with specified name.
	 * 
	 * @return Parameter value or <code>null</code> if such parameter doesn't exist.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Object getParameter(String name) {
		SWFuncParameterHome pHome = getFuncParameterHome();

		try {
			SWFuncParameter param = pHome.findByPrimaryKey(new SWFuncParameterPK(funcID, name));

			return param.getParameterValue();
		} catch (RemoteException e) {
			log.error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	/**
	 * Gets all <code>SWFuncColsMapping</code> instances for this functionality.
	 * 
	 * @return collection of <code>SWFuncColsMapping</code> instances for this functionality or <code>null</code> if
	 * there are no mappings for this functionality.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Collection getFuncColsMappings() {
		SWFuncColsMappingHome funcColsMappingHome = getFuncColsMappingHome();

		try {
			Collection res = funcColsMappingHome.findByCombination(funcID, getSWTableID(), null, null);

			return res;
		} catch (RemoteException e) {
			log.error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	/**
	 * Gets <code>SWFuncColsMapping</code> instance for this functionality with specified alias.
	 * 
	 * @param alias Mapping alias.
	 * 
	 * @return <code>SWFuncColsMapping</code> instance for this functionality with specified alias
	 * or <code>null</code> if there are no mappings for this functionality with such name.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public com.supportwizard.functionalities.interfaces.SWFuncColsMapping getFuncColsMapping(String alias) {

		// may be this is xdoclet's problem, it needs full specification of SWFuncColsMapping here
		SWFuncColsMappingHome funcColsMappingHome = getFuncColsMappingHome();

		try {
			Collection res = funcColsMappingHome.findByCombination(funcID, getSWTableID(), null, alias);

			log.assertLog(res.size() < 2,
					"there are more than one mapping for funcID=" + getFunctionalityID() + " with alias '" + alias + "'");

			if (res.size() == 0) {
				return null;
			}

			return (SWFuncColsMapping) (res.iterator().next());
		} catch (RemoteException e) {
			log.error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	/**
	 * Gets all functionality parameters.
	 * 
	 * @return Collection of <code>SWFuncParameter</code> instances or <code>null</code> if there are no parameters.
	 * 
	 * @ejb:interface-method view-type="remote"
	 * @ejb:transaction type="Supports"
	 */
	public Collection getParameters() {
		SWFuncParameterHome pHome = getFuncParameterHome();

		try {
			Collection res = pHome.findByFuncID(funcID);

			return res;
		} catch (RemoteException e) {
			log.error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	// some common ejb methods////////////////////////////////////////////////////////////
	public void setEntityContext(EntityContext context) {
		entityContext = context;
	}

	public void ejbLoad() {
		funcID = (Long) entityContext.getPrimaryKey();

		log.assertLog(funcID != null, "funcID==null in ejbLoad");

		SWInternal_SQL sw_sql = null;

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			return;
		}

		try {

			// select SWTableID, SWFuncMetaID, Enabled from SWFunctionalities where SWFuncID = id
			ColumnList columnList = new ColumnList();

			columnList.add("SWTableID");
			columnList.add("SWFuncMetaID");
			columnList.add("Enabled");

			TableList from = new TableList();

			from.add(TABLE_NAME, "");

			ConditionList where = new ConditionList();

			where.add("", "SWFuncID", "", "=", funcID, "");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			if (!rs.next()) {
				rs.close();

				throw new EJBException("Cant find SWFunctionality for SWFuncID = " + funcID);
			}

			long tableID = rs.getLong("SWTableID");

			funcMetaID = new Long(rs.getLong("SWFuncMetaID"));
			enabled = rs.getBoolean("Enabled");

			rs.close();

			SWTableHome tableHome = (SWTableHome) getHomesGetter().getHome(SWTableHome.class, "dictionary/SWTable");

			table = tableHome.findByPrimaryKey(new Long(tableID));

			if (table == null) {
				throw new EJBException("Can't find SWTable for SWTableID = " + tableID);

				// read name and description
			}

			columnList.clear();
			columnList.add("FuncName");
			columnList.add("FuncDescription");
			columnList.add("FuncHomeClassName");
			columnList.add("FuncHomeJndiName");
			from.clear();
			from.add("swfuncmetadata", "");
			where.clear();
			where.add("", "SWFuncMetaID", "", "=", funcMetaID, "");
			params.clear();
			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			rs = sw_sql.sqlQuery(params);

			if (!rs.next()) {
				rs.close();

				throw new EJBException("Cant find SWFuncMetadata for SWFuncMetaID = " + funcMetaID);
			}

			name = rs.getString("FuncName");
			description = rs.getString("FuncDescription");
			homeClassName = rs.getString("FuncHomeClassName");
			homeJndiName = rs.getString("FuncHomeJndiName");

			rs.close();
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}
	}

	public void ejbStore() {

		// save 'enabled' state
		SWInternal_SQL sw_sql = null;

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {

			// update SWFunctionalities set Enabled = enabled where SWFuncID = id
			ConditionList set = new ConditionList();

			set.add("", "Enabled", "", "", new Boolean(enabled), "");

			ConditionList where = new ConditionList();

			where.add("", "SWFuncID", "", "=", funcID, "");

			HashMap params = new HashMap();

			params.put("table", TABLE_NAME);
			params.put("set", set);
			params.put("where", where);

			if (sw_sql.sqlDML("update", params) == 0) {
				throw new Exception("Cant update " + TABLE_NAME + " for SWFuncID = " + funcID);
			}
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}
	}

	public void ejbActivate() {
		log = Logger.getLogger(getClass());
	}

	public void ejbPassivate() {
	}

	public void ejbRemove() throws RemoveException {
		log.debug("Removing functionality with SWFuncID=" + funcID);

		// get all parameters
		Collection funcParams = null;

		try {
			SWFuncParameterHome funcParameterHome = getFuncParameterHome();

			funcParams = funcParameterHome.findByFuncID(getFunctionalityID());
		} catch (FinderException e) {
			log.warn("Finder exception, skipping", e);

			funcParams = new ArrayList();  // to prevent NPE

			// skip it
		} catch (Throwable e1) {
			log.error("Exception caught", e1);

			throw new EJBException2(e1);
		} 

		// force ejbLoad call on all parameters
		try {
			for (Iterator it = funcParams.iterator(); it.hasNext(); ((SWFuncParameter) (it.next())).getParameterValue()) ;
		} catch (RemoteException e) {
			log.error("Exception caught", e);
		} 

		// get all funcColsMappings
		Collection funcMappings = null;

		try {
			SWFuncColsMappingHome funcColsMappingHome =
					(SWFuncColsMappingHome) getHomesGetter().getHome(SWFuncColsMappingHome.class);

			funcMappings = funcColsMappingHome.findByCombination(getFunctionalityID(), null, null, null);
		} catch (FinderException e) {
			log.warn("Finder exception, skipping", e);

			funcMappings = new ArrayList();  // to prevent NPE

			// skip it
		} catch (Throwable e1) {
			log.debug("Exception caught", e1);

			throw new EJBException2(e1);
		} 

		// force ejbLoad call on all mappings
		try {
			for (Iterator it = funcMappings.iterator(); it.hasNext(); ((SWFuncColsMapping) (it.next())).getAlias()) ;
		} catch (RemoteException e) {
			log.error("Exception caught", e);
		} 

		// get SWInternal_SQL
		SWInternal_SQL sw_sql = null;

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		} 

		// remove functionality record from swfunctionalities table
		try {
			ConditionList where = new ConditionList();

			log.assertLog(funcID != null, "funcID==null");
			where.add("", "SWFuncID", "", "=", funcID, "");

			HashMap params = new HashMap();

			params.put("table", TABLE_NAME);
			params.put("where", where);

			if (sw_sql.sqlDML("delete", params) == 0) {
				throw new Exception("Can't remove record from " + TABLE_NAME + " where SWFuncID=" + funcID);
			}
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);

				// throw new RemoveException(e.getMessage());
			}
		} 

		// remove SWFuncParameters
		for (Iterator it = funcParams.iterator(); it.hasNext();) {
			SWFuncParameter p = (SWFuncParameter) it.next();

			try {
				p.remove();
			} catch (RemoteException e) {
				log.error("Exception caught", e);

				// skip it
			} catch (RemoveException e) {
				log.debug("Exception caught", e);

				String pkSuffix = "";

				try {
					pkSuffix = " with pk=" + p.getPrimaryKey();
				} catch (RemoteException e1) {
					log.error("Can't get parameter primary key", e1);

					throw new RemoveException("Can't remove SWFuncParameter" + pkSuffix);
				}
			}
		} 

		// remove SWFuncColsMappings
		for (Iterator it = funcMappings.iterator(); it.hasNext();) {
			SWFuncColsMapping m = (SWFuncColsMapping) it.next();

			try {
				m.remove();
			} catch (RemoteException e) {
				log.error("Exception caught", e);

				// skip it
			} catch (RemoveException e) {
				log.debug("Exception caught", e);

				String pkSuffix = "";

				try {
					pkSuffix = "with pk=" + m.getPrimaryKey();
				} catch (RemoteException e1) {
					log.error("Can't get mapping primary key", e1);

					throw new RemoveException("Can't remove SWFuncColsMapping" + pkSuffix);
				}
			}
		}

		clearTableRelatedCaches();
	} 

	// common finder methods implementation//////////////////////////////////////////

	/**
	 * Finds functionality by primary key.
	 * @param funcID Primary key.
	 * @return Primary key.
	 */
	public Long ejbFindByPrimaryKey(Long funcID) throws FinderException {
		SWInternal_SQL sw_sql = null;

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {

			// check that row exists
			// select SWTableID, SWFuncMetaID from SWFunctionalities where SWFuncID = id
			ColumnList columnList = new ColumnList();

			columnList.add("SWTableID");
			columnList.add("SWFuncMetaID");

			TableList from = new TableList();

			from.add(TABLE_NAME, "");

			ConditionList where = new ConditionList();

			where.add("", "SWFuncID", "", "=", funcID, "");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			if (!rs.next()) {
				rs.close();

				throw new ObjectNotFoundException("Cant find SWFunctionality for SWFuncID = " + funcID);
			}

			long tableID = rs.getLong("SWTableID");

			funcMetaID = new Long(rs.getLong("SWFuncMetaID"));

			rs.close();

			SWTableHome tableHome = (SWTableHome) getHomesGetter().getHome(SWTableHome.class, "dictionary/SWTable");

			table = tableHome.findByPrimaryKey(new Long(tableID));

			if (table == null) {
				throw new FinderException("Can't find SWTable for SWTableID = " + tableID);

				// check that we can read name and description
			}

			columnList.clear();
			columnList.add("SWFuncMetaID");
			from.clear();
			from.add("swfuncmetadata", "");
			where.clear();
			where.add("", "SWFuncMetaID", "", "=", funcMetaID, "");
			params.clear();
			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			rs = sw_sql.sqlQuery(params);

			if (!rs.next()) {
				rs.close();

				throw new FinderException("Cant find SWFuncMetadata for SWFuncMetaID = " + funcMetaID);
			}

			rs.close();
		} catch (FinderException e) {
			log.debug("Exception caught", e);

			throw e;
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}

		makeDirty();  // state will be loaded by ejbLoad

		return funcID;
	}

	/**
	 * Finds all functionalities with the same funcMetaID. funcMetaID is calculated using
	 * <code>getDefaultFuncName()</code>,
	 * <code>getDefaultFuncDescription()</code>,
	 * <code>getDefaultFuncHomeClassName()</code>,
	 * <code>getDefaultFuncHomeJndiName()</code>.
	 * @return Collection of primary keys.
	 * @see SWFunctionalityBase#findFuncMetaID
	 */
	public Collection ejbFindAll() throws FinderException {
		ArrayList res = new ArrayList();
		SWInternal_SQL sw_sql = null;

		try {
			funcMetaID = findFuncMetaID(getCachedFunctionalityCreationInfo().funcName,
					getCachedFunctionalityCreationInfo().funcDescription,
					getCachedFunctionalityCreationInfo().funcHomeClassName,
					getCachedFunctionalityCreationInfo().funcHomeJndiName);
		} catch (ObjectNotFoundException e) {

			// there are no functionalities yet
			return res;
		} catch (Throwable e) {
			log.warn("Can't get functionality metadata ID", e);

			throw new FinderException("Can't get functionality metadata ID");
		}

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {

			// select SWFuncID from SWFunctionalities where SWFuncMetaID=funcMetaID
			ColumnList columnList = new ColumnList();

			columnList.add("SWFuncID");

			TableList from = new TableList();

			from.add(TABLE_NAME, "");

			ConditionList where = new ConditionList();

			where.add("", "SWFuncMetaID", "", "=", funcMetaID, "");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			while (rs.next()) {
				res.add(new Long(rs.getLong("SWFuncID")));
			} 

			// res.add(new SWHistoryFunctionalityPK(new Long(rs.getLong("SWFuncID"))));
			rs.close();
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}

		return res;
	}

	/**
	 * Finds all functionalities with the same funcMetaID and given SWTable. funcMetaID is calculated using
	 * <code>getDefaultFuncName()</code>,
	 * <code>getDefaultFuncDescription()</code>,
	 * <code>getDefaultFuncHomeClassName()</code>,
	 * <code>getDefaultFuncHomeJndiName()</code>.
	 * @param table SWTable found functionalities must belong to.
	 * @return Collection of primary keys.
	 * @see SWFunctionalityBase#findFuncMetaID
	 */
	public Collection ejbFindByTable(SWTable table) throws FinderException {
		ArrayList res = new ArrayList();
		SWInternal_SQL sw_sql = null;

		try {
			funcMetaID = findFuncMetaID(getCachedFunctionalityCreationInfo().funcName,
					getCachedFunctionalityCreationInfo().funcDescription,
					getCachedFunctionalityCreationInfo().funcHomeClassName,
					getCachedFunctionalityCreationInfo().funcHomeJndiName);
		} catch (ObjectNotFoundException e) {

			// there are no functionalities yet
			return res;
		} catch (Throwable e) {
			log.warn("Can't get functionality metadata ID", e);

			throw new FinderException("Can't get functionality metadata ID");
		}

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {
			Long tableID = table.getSWTableID();

			// select SWFuncID from SWFunctionalities
			ColumnList columnList = new ColumnList();

			columnList.add("SWFuncID");

			TableList from = new TableList();

			from.add(TABLE_NAME, "");

			ConditionList where = new ConditionList();

			where.add("", "SWTableID", "", "=", tableID, "");
			where.add("", "SWFuncMetaID", "", "=", funcMetaID, "AND");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			while (rs.next()) {
				res.add(new Long(rs.getLong("SWFuncID")));
			} 

			// res.add(new SWHistoryFunctionalityPK(new Long(rs.getLong("SWFuncID"))));
			rs.close();
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		} 

		// try {
		// log.debug("SWFunctionalityBase.findByTable (" + funcMetaID + ", " + table.getName() + ") returning " + res.size() + " element" + (res.size() == 1?"":"s"));
		// } catch (RemoteException e) {
		// }
		return res;
	}

	/**
	 * Finds all enabled functionalities with the same funcMetaID and given SWTable. funcMetaID is calculated using
	 * <code>getDefaultFuncName()</code>,
	 * <code>getDefaultFuncDescription()</code>,
	 * <code>getDefaultFuncHomeClassName()</code>,
	 * <code>getDefaultFuncHomeJndiName()</code>.
	 * @param table SWTable found functionalities must belong to.
	 * @return Collection of primary keys.
	 * @see SWFunctionalityBase#findFuncMetaID
	 */
	public Collection ejbFindEnabledByTable(SWTable table) throws FinderException {
		ArrayList res = new ArrayList();
		SWInternal_SQL sw_sql = null;

		try {
			funcMetaID = findFuncMetaID(getCachedFunctionalityCreationInfo().funcName,
					getCachedFunctionalityCreationInfo().funcDescription,
					getCachedFunctionalityCreationInfo().funcHomeClassName,
					getCachedFunctionalityCreationInfo().funcHomeJndiName);
		} catch (ObjectNotFoundException e) {

			// there are no functionalities yet
			return res;
		} catch (Throwable e) {
			log.warn("Can't get functionality metadata ID", e);

			throw new FinderException("Can't get functionality metadata ID");
		}

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {
			Long tableID = table.getSWTableID();

			// select SWFuncID from SWFunctionalities
			ColumnList columnList = new ColumnList();

			columnList.add("SWFuncID");

			TableList from = new TableList();

			from.add(TABLE_NAME, "");

			ConditionList where = new ConditionList();

			where.add("", "SWTableID", "", "=", tableID, "");
			where.add("", "SWFuncMetaID", "", "=", funcMetaID, "AND");
			where.add("", "Enabled", "", "=", new Boolean(true), "AND");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columnList);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			while (rs.next()) {
				res.add(new Long(rs.getLong("SWFuncID")));
			} 

			// res.add(new SWHistoryFunctionalityPK(new Long(rs.getLong("SWFuncID"))));
			rs.close();
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		} 

		// try {
		// log.debug("SWFunctionalityBase.findEnabledByTable (" + funcMetaID + ", " + table.getName() + ") returning " + res.size() + " elements");
		// } catch (RemoteException e) {
		// }
		return res;
	}

	/**
	 * Creates new functionaliy associated with given table.
	 * @param seance Current user's seance. Used to get <code>HomesGetter</code> and localize error messages.
	 * @param table New functionality will belong to.
	 * @return Primary key.
	 * 
	 * ejb:create-method
	 * ejb:transaction type="Required"
	 */
	public Long ejbCreate(Seance seance, SWTable table) throws CreateException {
		if (seance != null) {
			homesGetter = seance.getHomesGetter();
		}

		Long res = create(getCachedFunctionalityCreationInfo().funcName,
				getCachedFunctionalityCreationInfo().funcDescription,
				getCachedFunctionalityCreationInfo().funcHomeClassName,
				getCachedFunctionalityCreationInfo().funcHomeJndiName, table, true);
		SWFuncColsMappingHome mappingHome = null;

		try {
			mappingHome = (SWFuncColsMappingHome) getHomesGetter().getHome(SWFuncColsMappingHome.class);
		} catch (NamingException e) {
			log.debug("JNDI problems", e);

			throw new CreateException(e.getMessage());
		} catch (Throwable e) {
			log.fatal("Problems..", e);

			throw new CreateException(e.getMessage());
		} 

		// create all parameters
		int nParams = getCachedFunctionalityCreationInfo().getNumberOfParameters();

		if (nParams != 0) {
			log.debug("attempting to create " + nParams + " parameter(s)");

			for (int i = 0; i < nParams; i++) {
				SWFuncParameterCreationInfo paramCreationInfo;

				paramCreationInfo = getCachedFunctionalityCreationInfo().getParameterCreationInfo(i);

				try {
					log.debug("Creating parameter with paramCreationInfo=" + paramCreationInfo);
					getFuncParameterHome().create(funcID, funcMetaID, paramCreationInfo.paramName, paramCreationInfo.paramDescription,
							paramCreationInfo.paramDefaultValue);
					log.debug("Parameter #" + i + ", '" + paramCreationInfo.paramName + "' successfully created");
				} catch (Throwable e) {
					log.error("Can't create functionality parameter", e);

					throw new CreateException("Can't create functionality parameter");
				}
			}
		} 

		// create all column mappings
		int nMappings = getCachedFunctionalityCreationInfo().getNumberOfMappings();

		if (nMappings != 0) {
			log.debug("attempting to create " + nMappings + " mapping(s)");

			for (int i = 0; i < nMappings; i++) {
				SWFuncColsMappingCreationInfo mappingCreationInfo;

				mappingCreationInfo = getCachedFunctionalityCreationInfo().getMappingCreationInfo(i);

				try {
					log.debug("Creating column mapping with mappingCreationInfo=" + mappingCreationInfo);
					mappingHome.create(funcID, getSWTableID(), mappingCreationInfo.columnID, mappingCreationInfo.alias);
					log.debug("Mapping #" + i + ", '" + mappingCreationInfo.alias + "' successfully created");
				} catch (Throwable e) {
					log.error("Can't create functionality column mapping", e);

					throw new CreateException("Can't create functionality column mapping");
				}
			}
		}

		clearTableRelatedCaches();

		return res;
	}

	/**
	 * Creates new functionaliy associated with given table.
	 * @param table New functionality will belong to.
	 * @return Primary key.
	 * 
	 * @ejb:create-method
	 * @ejb:transaction type="Required"
	 */
	public Long ejbCreate(SWTable table) throws CreateException {
		return ejbCreate(null, table);
	}

	public void ejbPostCreate(SWTable table) {
	}

	public void ejbPostCreate(Seance seance, SWTable table) {
	}

	/**
	 * This is a helper method for writing ejbCreate methods. First it tries to find suitable functionality
	 * metadata and reads it's SWFuncMetaID. If nothing can be found then new record in SWFuncMetadata is created.
	 * Then this method adds new record into SWFunctionalities table.
	 * 
	 * @param funcName The functionalitiy's name. Must appear in FuncName field of SWFunMetadata table.
	 * @param funcDescription The functionalitiy's description. Must appear in FuncDescription field of SWFunMetadata table.
	 * @param funcHomeClassName The functionalitiy's home interface class name. Must appear in FuncHomeClassName field of SWFunMetadata table.
	 * @param funcHomeClassName The functionalitiy's home interface jndi name. Must appear in FuncHomeJndiName field of SWFunMetadata table.
	 * @param table The table functionalitiy must be connected with. It's ID will be inserted into SWTableID field of SWFunctionalities table.
	 * @param enabled Initial functionality state.
	 * 
	 * @return Primary key.
	 */
	protected Long create(String funcName, String funcDescription, String funcHomeClassName, String funcHomeJndiName,
						  SWTable table, boolean enabled) throws CreateException {
		log.debug("Creating functionality with funcName='" + funcName + "', funcDescription='" + funcDescription
				+ "', funcHomeClassName='" + funcHomeClassName + "', funcHomeJndiName='" + funcHomeJndiName + '\'');

		name = funcName;
		description = funcDescription;
		homeClassName = funcHomeClassName;
		homeJndiName = funcHomeJndiName;
		this.enabled = enabled;
		this.table = table;

		SWInternal_SQL sw_sql = null;

		try {
			sw_sql = new SWInternal_SQL();

			sw_sql.dbConnect();
		} catch (Throwable e) {
			log.error("Can't connect to the database", e);

			throw new EJBException2(e);
		}

		try {

			// first try to find suitable metadata
			ColumnList columns = new ColumnList();

			columns.add("SWFuncMetaID");

			TableList from = new TableList();

			from.add("swfuncmetadata", "");

			ConditionList where = new ConditionList();

			where.add("", "FuncName", "", "=", funcName, "");
			where.add("", "FuncDescription", "", "=", funcDescription, "AND");
			where.add("", "FuncHomeClassName", "", "=", funcHomeClassName, "AND");
			where.add("", "FuncHomeJndiName", "", "=", funcHomeJndiName, "AND");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columns);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			if (rs.next()) {
				log.debug("Found suitable metadata");

				funcMetaID = new Long(rs.getLong("SWFuncMetaID"));

				rs.close();
			} else {
				log.debug("Clearing old metadata");
				params.put("table", "swfuncmetadata");
				where.clear();
				where.add("", "FuncName", "", "=", funcName, "");
				params.put("where", where);  // I know this statement is redundant
				sw_sql.sqlDML("delete", params);
				log.debug("Creating new metadata");
				rs.close();
				columns.clear();
				columns.add("FuncName");
				columns.add("FuncDescription");
				columns.add("FuncHomeClassName");
				columns.add("FuncHomeJndiName");

				ArrayList values = new ArrayList();

				values.add(funcName);
				values.add(funcDescription);
				values.add(funcHomeClassName);
				values.add(funcHomeJndiName);
				params.clear();
				params.put("table", "swfuncmetadata");
				params.put("columns", columns);
				params.put("values", values);

				if (sw_sql.sqlDML("insert", params) == 0) {
					throw new Exception("Can't create functionality metadata record");
				}

				int id = sw_sql.getInsertID();

				funcMetaID = new Long(id);
			} 

			// now we have metadata id and we must add new record to SWFunctionalities
			columns.clear();
			columns.add("SWtableID");
			columns.add("SWFuncMetaID");
			columns.add("Enabled");

			ArrayList values = new ArrayList();

			values.add(table.getSWTableID());
			values.add(funcMetaID);
			values.add(new Boolean(enabled));
			params.clear();
			params.put("table", TABLE_NAME);
			params.put("columns", columns);
			params.put("values", values);

			if (sw_sql.sqlDML("insert", params) == 0) {
				throw new Exception("Can't create functionality record");
			}

			int id = sw_sql.getInsertID();

			funcID = new Long(id);

			log.debug("Created functionality with funcID=" + funcID);
		} catch (CreateException e) {
			log.error("Exception caught", e);

			throw e;
		} catch (Throwable e) {
			log.error("Exception caught", e);

			throw new EJBException2(e);
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}

		log.debug("Functionality " + funcName + " has been successfully created");

		return funcID;
	}

	/**
	 * Tries to find funcMetaID by given fields.
	 * Uses following SQL:
	 * <code>
	 * select SWFuncMetaID from SWFuncMetadata where FuncName=funcName and FuncDescription=funcDescription and FuncHomeClassName=funcHomeClassName and FuncHomeJndiName=funcHomeJndiName
	 * </code>
	 * @param funcName The functionalitiy's name. Must appear in FuncName field of SWFunMetadata table.
	 * @param funcDescription The functionalitiy's description. Must appear in FuncDescription field of SWFunMetadata table.
	 * @param funcHomeClassName The functionalitiy's home interface class name. Must appear in FuncHomeClassName field of SWFunMetadata table.
	 * @param funcHomeClassName The functionalitiy's home interface jndi name. Must appear in FuncHomeJndiName field of SWFunMetadata table.
	 * @return Functionality metadata ID.
	 * 
	 * @throws FinderException if functionality metadata doesn't exist.
	 * @throws SQLException if there is an error during results retrieval.
	 * @throws SW_SQLException if there's an error during DB connecting.
	 */
	protected Long findFuncMetaID(String funcName, String funcDescription, String funcHomeClassName,
								  String funcHomeJndiName) throws FinderException, SQLException, SW_SQLException {
		SWInternal_SQL sw_sql = null;
		Long res;

		sw_sql = new SWInternal_SQL();

		sw_sql.dbConnect();

		try {

			// SQL: select SWFuncMetaID from SWFuncMetadata where FuncName=funcName and FuncDescription=funcDescription and FuncHomeClassName=funcHomeClassName and FuncHomeJndiName=funcHomeJndiName
			ColumnList columns = new ColumnList();

			columns.add("SWFuncMetaID");

			TableList from = new TableList();

			from.add("swfuncmetadata", "");

			ConditionList where = new ConditionList();

			where.add("", "FuncName", "", "=", funcName, "");
			where.add("", "FuncDescription", "", "=", funcDescription, "AND");
			where.add("", "FuncHomeClassName", "", "=", funcHomeClassName, "AND");
			where.add("", "FuncHomeJndiName", "", "=", funcHomeJndiName, "AND");

			HashMap params = new HashMap();

			params.put("from", from);
			params.put("columns", columns);
			params.put("where", where);

			ResultSet rs = sw_sql.sqlQuery(params);

			if (rs.next()) {
				res = new Long(rs.getLong("SWFuncMetaID"));

				rs.close();
			} else {
				throw new ObjectNotFoundException("Can't find functionality metadata for where=" + where);
			}
		} finally {
			try {
				sw_sql.dbDisconnect();
			} catch (SQLException e) {
				log.error("Cant't close database connection", e);
			}
		}

		return res;
	} 

	// /////////////////////////////

	/**
	 * Gets <code>SWFuncParameterHome</code>.
	 * 
	 * @return <code>SWFuncParameterHome</code>.
	 */
	protected SWFuncParameterHome getFuncParameterHome() {
		try {
			return (SWFuncParameterHome) getHomesGetter().getHome(SWFuncParameterHome.class);
		} catch (Throwable e) {
			log.error("Can't get SWFuncParameterHome", e);

			throw new EJBException2(e);
		}
	}

	/**
	 * Gets <code>SWFuncColsMappingHome</code>.
	 * 
	 * @return <code>SWFuncColsMappingHome</code>.
	 */
	protected SWFuncColsMappingHome getFuncColsMappingHome() {
		try {
			return (SWFuncColsMappingHome) getHomesGetter().getHome(SWFuncColsMappingHome.class);
		} catch (Throwable e) {
			log.error("Can't get SWFuncParameterHome", e);

			throw new EJBException2(e);
		}
	}

	/**
	 * Gets initialized simple cache implementation.
	 * 
	 * @return initialized simple cache implementation.
	 */
	protected SimpleCache getCache() {
		if (cache != null) {
			return cache;
		}

		cache = SimpleCacheFactory.getSimpleCache();

		return cache;
	}

	/**
	 * Gets table-related cache wrapper.
	 * 
	 * @return table-related cache wrapper.
	 */
	public SWTableRelatedCachesWrapper getTableRelatedCachesWrapper() {
		if (tableRelatedCachesWrapper != null) {
			return tableRelatedCachesWrapper;
		}

		tableRelatedCachesWrapper = new SWTableRelatedCachesWrapper();

		return tableRelatedCachesWrapper;
	}

	/**
	 * Returns a homes getter.
	 * @return Not <code>null</code> homes getter.
	 */
	protected HomesGetter getHomesGetter() {
		try {
			if (homesGetter == null) {
				homesGetter = new HomesGetter();
			}
		} catch (NamingException e) {
			log.debug("Exception caught", e);

			throw new EJBException(e);
		}

		return homesGetter;
	}

	/**
	 * Sets homes getter.
	 * 
	 * @param homesGetter homes getter.
	 */
	protected void setHomesGetter(HomesGetter homesGetter) {
		this.homesGetter = homesGetter;
	}

	// ////////////////////////////////// private area
	private SWFunctionalityCreationInfo getCachedFunctionalityCreationInfo() {
		if (functionalityCreationInfo == null) {
			functionalityCreationInfo = getFunctionalityCreationInfo();
		}

		return functionalityCreationInfo;
	}

	private void clearTableRelatedCaches() {
		getTableRelatedCachesWrapper().clean(getSWTableID());
	}

}

