/*
 * SupportWizard 2
 *
 * Copyright (C) 2001 ISC corp. All Rights Reserved.
 *
 * $Id: SWFunctionalityLogic.java,v 1.31 2003/08/30 18:03:32 kos Exp $
 */

package com.supportwizard.functionalities;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Map;
import java.util.List;

import com.supportwizard.dictionary.SWColumn;
import com.supportwizard.dictionary.SWColumnValue;
import com.supportwizard.dictionary.SWTable;
import com.supportwizard.dml.interfaces.SWRecord;
import com.supportwizard.dml.SWData;
import com.supportwizard.seance.Seanceable;
import com.supportwizard.db.ConditionList;

/**
 * This interface contains all functionality callbacks. All methods that work with column name-&gt;field
 * value mapping work with <strong>user</strong> column names ({@link SWColumnValue#getName}),
 * not with the DB-names ({@link SWColumnValue#getDBName})<p/>
 *
 * Use {@link com.supportwizard.functionalities.SWFunctionalityLogicBase SWFunctionalityLogicBase}
 * as superclass when you don't need to handle all events.
 *
 * @author <a href="mailto:kos@supportwizard.com">Konstantin Sobolev</a>
 * @version $Revision: 1.31 $
 */
public interface SWFunctionalityLogic extends Seanceable, Serializable{
//following beans are used (ejb-ref's needed):
//SWColumn, SWRecord, SWData, Seance

    /**
     * Called after the new column has been added.
     *
     * @param column Column that has been added.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dictionary.SWColumnBean
     */
    void columnAdded(SWColumn column) throws SWFunctionalityException, RemoteException;

    /**
     * Called when the column is about to be removed.
     *
     * @param column Column that is about to be removed.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dictionary.SWColumnBean
     */
    void columnRemoving(SWColumn column) throws SWFunctionalityException, RemoteException;

    /**
     * Called when new record is about to be added.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param recordData Record (column name - field value pairs) that is about to be added.
     * @return possibly modified record data.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    Map recordAdding(int operationInitiator, Map recordData) throws SWFunctionalityException, RemoteException;

    /**
     * Called when new record has been added.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that has been added.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    void recordAdded(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException;

    /**
     * Called when table record is about to be modified.
     * At this point old values can still be obtained using
     * <code>record.getValue()</code> call.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that is about to be modified.
     * @param newValues New record values.
     * @return possibly modified new values.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    Map recordModifying(int operationInitiator, SWRecord record, Map newValues) throws SWFunctionalityException, RemoteException;

    /**
     * Called when table record has been modified.
     * At this point <code>record</code> already contains new values and can't be modified.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that has been modified.
     * @param oldValues Old record values.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    void recordModified(int operationInitiator, SWRecord record, Map oldValues) throws SWFunctionalityException, RemoteException;

    /**
     * Called when the record is about to be removed.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that is about to be removed.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    void recordRemoving(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException;

	/**
     * Called when the record has been removed.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param recordData data from the removed record.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     */
	void recordRemoved(int operationInitiator, Map recordData) throws SWFunctionalityException, RemoteException;

    /**
     * Called when new record is about to be imported. The implementations must work as fast as possible.
     *
     * @param recordData Record (column name - field value pairs) that is about to be imported.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     * @return possibly modified record values.
     */
    Map recordImporting(Map recordData) throws SWFunctionalityException, RemoteException;

    /**
     * Called when new record has been imported. The implementations must work as fast as possible.
     *
     * @param record Record that has been imported.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    void recordImported(SWRecord record) throws SWFunctionalityException, RemoteException;

    /**
     * Called at the end of the importing process, when all records were added.
     *
     * @throws SWFunctionalityException If something was wrong during event handling.
     * @see com.supportwizard.importer.XMLImporter
     */
    void importFinished() throws SWFunctionalityException, RemoteException;

    /**
     * Called when the table is about to be renamed.
     * At this point old table name can still be obtained using
     * <code>table.getName()</code> call (Each functinality must
     * know about the table it is attached to).
     *
     * @param newName New table name.
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     */
    void tableRenaming(String newName) throws SWFunctionalityException, RemoteException;

    //modification checking callbacks///////////////////////////////

    /**
     * Checks wether column with specified value object may be added.
     *
     * @return true if <code>column</code> may be added; false otherwise.
     *
     * @see com.supportwizard.dictionary.SWColumnBean
     */
    boolean isColumnAdditionAllowed(SWColumnValue columnValue) throws RemoteException;

    /**
     * Checks wether specified column may be removed.
     *
     * @return true if <code>column</code> may be removed; false otherwise.
     *
     * @see com.supportwizard.dictionary.SWColumnBean
     */
    boolean isColumnRemovalAllowed(SWColumn column) throws RemoteException;

    /**
     * Checks wether specified record may be added.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param recordData Record (column name - field value pairs) that is about to be added.
     *
     * @return true if record containing <code>recordData</code> may be added; false otherwise.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    boolean isRecordAdditionAllowed(int operationInitiator, Map recordData) throws RemoteException, SWFunctionalityException;

    /**
     * Checks wether specified record may be removed.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that is about to be removed.
     *
     * @return true if <code>record</code> may be removed; false otherwise.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    boolean isRecordRemovalAllowed(int operationInitiator, SWRecord record) throws RemoteException, SWFunctionalityException;

    /**
     * Checks wether specified record may be modified.
     *
     * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
     * see {@link SWFunctionalitiesConstants}
     * @param record Record that is about to be modified.
     * @param newValues New record values.
     *
     * @return true if <code>record</code> may be modified; false otherwise.
     *
     * @see com.supportwizard.dml.ejb.SWRecordBean
     */
    boolean isRecordModificationAllowed(int operationInitiator, SWRecord record, Map newValues) throws RemoteException, SWFunctionalityException;

    /**
     * Checks wether specified table may be renamed.
     *
     * @return true if table may be renamed to <code>newName</code>; false otherwise.
     */
    boolean isTableRenamingAllowed(String newName) throws RemoteException;

     /**
     * Checks wether specified records may be imported.
     *
      * @param data data that is about to be imported. list of maps.
     *
     * @return true if all of <code>data</code> rows are allowed to be imported.
     */
    boolean isRecordsImportAllowed(List data) throws RemoteException, SWFunctionalityException;

    /**
     * Called on reading read-only data. Usually used for displaying the whole table
     * in red-only mode. Thus it is unnecessary to fill in any additional SWData's
     * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
     * However, functionality logic may modify some fields or even add virtual fields.
     *
     * @param data Data being read. All functionalities will in series receive the same SWData instance.
     * @return Data modified by this callback.
     *
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     */
    SWData roDataReading (SWData data) throws SWFunctionalityException, RemoteException;

    /**
     * Called on reading some database data that may be modified by user. Usually used for displaying
     * and modifying single line from the table. At this point functionality logic can insert some additional
     * info into SWData's
     * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
     * Functionality logic can aslo modify some fields or add virtual fields.
     *
     * @param data Data being read. All functionalities will in series receive the same SWData instance.
     * @return Data modified by this callback.
     *
     * @throws com.supportwizard.functionalities.SWFunctionalityException If something
     * was wrong during event handling.
     */
    SWData rwDataReading (SWData data) throws SWFunctionalityException, RemoteException;

    /**
     * This callback should eliminate condition items based on artificial columns.
     * It is impossible to directly query SW tables using artificial fiedls in WHERE clause;
     * Functionality should implement this method, if it maintains some artificial fields and
     * there's reasonable chance that client will use this field in ConditionList.
     * <p>
     * There are three options on how to maintain artificial fields in ConditionList:
     * <ul>
     * <li>Functionality can throw <code>SWFunctionalityException</code> if it does not want anybody
     * to filter by particular artificial fiedl. It is acceptable, when filtering by some artificial field does not
     * make sence or if it is very difficult to implement (filtering by history columns seems a good example here).
     * Another example you can find in implementation of <tt>SWContactFunctionalityLogic</tt>, where functionality
     * throws an exception when detects a constraint based on <tt>password</tt> field.
     * </li>
     * <li>Functionality can transform query parameters in such a way that they will not contain
     * artificial constraints; it can substitute artificial filters with filters by non-artificial
     * fields (see <tt>SWContacsFunctionalityLogic</tt> as possible example, where filtering by <tt>
     * login</tt> field is substituted by filtering by main table's primary key). <b>Exercise caution
     * when implementing this approach in a callback</b>. Make sure that <tt>ConditionList</tt> is still
     * in consistent state after your modifications.
     * </li>
     * </ul>
     * It is not recommended to simply remove <tt>ConditionElement</tt>s from <tt>ConditionList</tt>,
     * without providing of an appropriate substitution with physical column filter.
     * @param table Table we're working with.
     * @param filters Query filters in the form of <tt>ConditionList</tt>. Functionality can modify this list.
     * @throws SWFunctionalityException If evil comes
     */
    public void prepareVirtualConstraints(SWTable table, ConditionList filters) throws RemoteException, SWFunctionalityException;
    /**
     * Gets an array of names of tables that must be dropped when functionality is removed.
     * Here we mean not <code>SWTable</code> but simple database tables. In case if functionality
     * wants to kill <code>SWTable</code> it must invoke <code>SWTable.remove()</code> itself
     * (for example it it's own <code>ejbRemove()</code>) and return table's name here.
     * <p>
     * This method helps to create more or less safe way to remove functionalities. The problem is in
     * transactions handling: mySQL (and most other RDBMS) can't rollback DDL commands (create table, drop
     * table, alter table). Thus there can occur a situation when dictionary removes a number of fuctionalities
     * associated with some table before removing the table itself and these functionalities wants to drop
     * other tables. If funtionalities will invoke <code>SWInternalSQL.qlDDL(...)</code> themselves and
     * some functionality will encounter an allpication-level exception then the overall transaction
     * cant' be correctly rolled back and this will lead to DB inconsistency.
     * <p>
     * The solution is to execute all DDL instructions at the end of transaction. In this case all apllication
     * level exceptions can be safely processed and transaction can be rolled back.
     *
     * @return Array of names of tables that must be dropped when functionality is removed or <code>null</code>
     * as synonym of empty list.
     */
    String[] getTablesToDropOnRemove () throws RemoteException;

    /**
     * Sets functionality logic bus. See {@link SWFunctionalityLogicBus} documentation
     * for details.
     *
     * @param bus Functionality logic bus.
     */
    void setBus (SWFunctionalityLogicBus bus) throws RemoteException;
}