/*
 * SupportWizard 2
 *
 * Copyright (C) 2001 ISC corp. All Rights Reserved.
 *
 * $Id: SWFunctionalityLogicBase.java,v 1.46 2003/08/30 18:03:32 kos Exp $
 */
package com.supportwizard.functionalities;

import com.supportwizard.dictionary.SWColumn;
import com.supportwizard.dictionary.SWColumnValue;
import com.supportwizard.dictionary.SWTable;
import com.supportwizard.dml.SWData;
import com.supportwizard.dml.ejb.SWRecordBean;
import com.supportwizard.dml.interfaces.SWRecord;
import com.supportwizard.functionalities.interfaces.*;
import com.supportwizard.seance.Seance;
import com.supportwizard.utils.HomesGetter;
import com.supportwizard.cache.SimpleCache;
import com.supportwizard.cache.SimpleCacheFactory;
import com.supportwizard.db.ConditionList;
import org.apache.log4j.Logger;

import javax.ejb.FinderException;
import javax.naming.NamingException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Map;
import java.util.List;

// beans used (ejb-ref's needed):
// SWColumn, SWRecord, SWData, Seance

/**
 * Proposed stub for functionality logic implementations. Contains empty
 * methods. When subclassing use {@link com.supportwizard.functionalities.SWFunctionalityLogicBase#getLog() log()}
 * method for getting logger instance.
 *
 * @author <a href="mailto:kos@supportwizard.com">Konstantin Sobolev</a>
 * @version $Revision: 1.46 $
 */
public class SWFunctionalityLogicBase implements SWFunctionalityLogic {
	private static final boolean AUTO_CHECK_PERMISSIONS = true;

	/**
	 * Logging category; should be accessed by #getLog().
	 */
	private transient Logger log;
	private transient SimpleCache cache;

	/**
	 * Seance; may be used by derived classes.
	 */
	protected Seance seance = null;

	/**
	 * Functionality ID.
	 */
	protected Long funcID;

	/**
	 * User's table ID. (table the functionality is associated with.
	 */
	protected Long tableID;

	/**
	 * Homes getter; may be used by derived classes.
	 */
	protected HomesGetter homesGetter;

	/**
	 * Functionality logic bus.
	 */
	protected SWFunctionalityLogicBus bus;

	/**
	 * Basic constructor. To be called by children.
	 */
	protected SWFunctionalityLogicBase(Long funcID, Long tableID) {
		this.funcID = funcID;
		this.tableID = tableID;

		try {
			if (seance != null)
				homesGetter = seance.getHomesGetter();
			else {
				// log().debug("seance isn't set"); //in most cases it can't be set here
				homesGetter = new HomesGetter();
			}
		} catch (NamingException e) {
			getLog().fatal("Unexpected exception", e);
		}
	}

	// seanceable implementation
	public void setSeance(Seance seance) {
		if (seance == null) {
			getLog().error("Attempt to set null seance! Ignoring..");

			if (getLog().isDebugEnabled())
				getLog().debug("Attempt to set null seance", new Exception("debug stacktrace"));

			return;
		}

		this.seance = seance;
		homesGetter = seance.getHomesGetter();
	}

	public Seance getSeance() {
		if (seance == null) {
			getLog().warn("Seance is not set");

			if (getLog().isDebugEnabled())
				getLog().debug("Seance is not set", new Exception("debug stacktrace"));
		}

		return seance;
	}

	// some helper methods

	/**
	 * Gets functionality parameter with specified name.
	 *
	 * @return Parameter value or <code>null</code> if such parameter doesn't exist.
	 */
	public String getParameter(String name) {
		SWFuncParameterHome pHome = getFuncParameterHome();

		try {
			SWFuncParameter param = pHome.findByPrimaryKey(new SWFuncParameterPK(funcID, name));

			return param.getParameterValue();
		} catch (RemoteException e) {
			getLog().error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	public void setParameter(String name, String description, String value) {
		SWFuncParameterHome pHome = getFuncParameterHome();
		SWFunctionalitiesFinderHome finderHome = getFunctionalitiesFinderHome();

		try {
			SWFunctionalitiesFinder finder = finderHome.create();
			SWFunctionality functionality = finder.findByFunctionalityID(funcID);

			finder.remove();
			pHome.create(functionality, name, description, value);
		} catch (Exception e) {
			getLog().error("Exception caught", e);
		}
	}

	/**
	 * Gets all <code>SWFuncColsMapping</code> instances for this functionality.
	 *
	 * @return collection of <code>SWFuncColsMapping</code> instances for this functionality or <code>null</code> if
	 * there are no mappings for this functionality.
	 */
	public Collection getFuncColsMappings() {
		SWFuncColsMappingHome _funcColsMappingHome = getFuncColsMappingHome();

		try {
			Collection res = _funcColsMappingHome.findByCombination(funcID, tableID, null, null);

			return res;
		} catch (RemoteException e) {
			getLog().error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	protected SWFunctionalitiesFinderHome getFunctionalitiesFinderHome() {
		try {
			return (SWFunctionalitiesFinderHome) homesGetter.getHome(SWFunctionalitiesFinderHome.class);
		} catch (Throwable e) {
			getLog().error("Can't get SWFunctionalitiesFinderHome", e);

			return null;
		}
	}

	protected SWFuncParameterHome getFuncParameterHome() {
		try {
			return (SWFuncParameterHome) homesGetter.getHome(SWFuncParameterHome.class);
		} catch (Throwable e) {
			getLog().error("Can't get SWFuncParameterHome", e);

			return null;
		}
	}

	protected SWFuncColsMappingHome getFuncColsMappingHome() {
		try {
			return (SWFuncColsMappingHome) homesGetter.getHome(SWFuncColsMappingHome.class);
		} catch (Throwable e) {
			getLog().error("Can't get getFuncColsMappingHome", e);

			return null;
		}
	}

	/**
	 * Gets all functionality parameters.
	 *
	 * @return Collection of <code>SWFuncParameter</code> instances or <code>null</code> if there are no parameters.
	 */
	public Collection getParameters() {
		SWFuncParameterHome pHome = getFuncParameterHome();

		try {
			Collection res = pHome.findByFuncID(funcID);

			return res;
		} catch (RemoteException e) {
			getLog().error("Exception caught", e);

			return null;
		} catch (FinderException e) {
			return null;
		}
	}

	// callbacks implementations

	/**
	 * Called when new column was added.
	 *
	 * @param column Column that is about to be removed.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void columnAdded(SWColumn column) throws SWFunctionalityException, RemoteException {
		// if (AUTO_CHECK_PERMISSIONS)
		// try {
		// if (!isColumnAdditionAllowed(column.getColumnValue()))
		// throw new SWFunctionalityException("Column addition is not allowed");
		// } catch (RemoteException e) {
		// getLog().error("Exception caught", e);
		// }
	}

	/**
	 * Called when the column is about to be removed.
	 *
	 * @param column Column that is about to be removed.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void columnRemoving(SWColumn column) throws SWFunctionalityException, RemoteException {
		if (AUTO_CHECK_PERMISSIONS)
			if (!isColumnRemovalAllowed(column))
				throw new SWFunctionalityException("Column removal is not allowed");
	}

	/**
	 * Called when new record is about to be added.
	 *
	 * @param recordData Record (column name - field value pairs) that is about to be added.
	 * @return possibly modified record data.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 *
	 * @see com.supportwizard.dml.ejb.SWRecordBean
	 */
	public Map recordAdding(int operationInitiator, Map recordData) throws SWFunctionalityException, RemoteException {
		if (AUTO_CHECK_PERMISSIONS)
			if (!isRecordAdditionAllowed(operationInitiator, recordData))
				throw new SWFunctionalityException("Record addition as not allowed");

		return recordData;
	}

	/**
	 * Called when new record was added.
	 *
	 * @param record Record that was added.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void recordAdded(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException {
	}

	/**
	 * Called when new record is about to be imported. The implementations must work as fast as possible.
	 *
	 * @param recordData Record (column name - field value pairs) that is about to be imported.
	 * @throws SWFunctionalityException If something
	 * was wrong during event handling.
	 * @return possibly modified record values.
	 */
	public Map recordImporting(Map recordData) throws SWFunctionalityException, RemoteException {
		return recordAdding(SWFunctionalitiesConstants.BY_API, recordData);
	}

	/**
	 * Called when new record was imported. This callback was introduced as a replacement for
	 * {@link #isRecordAdditionAllowed} {@link #recordAdding} {@link #recordAdded} sequence when record is being
	 * imported, mostly to increase performance. The implementations must work as fast as possible.
	 *
	 * @param record Record that was imported.
	 * @throws SWFunctionalityException If something
	 * was wrong during event handling.
	 *
	 * @see SWRecordBean
	 */
	public void recordImported(SWRecord record) throws SWFunctionalityException, RemoteException {
		recordAdded(SWFunctionalitiesConstants.BY_API, record);
	}

	/**
	 * Called at the end of the importing process, when all records were added.
	 *
	 * @throws SWFunctionalityException If something was wrong during event handling.
	 * @see com.supportwizard.importer.XMLImporter
	 */
	public void importFinished() throws SWFunctionalityException, RemoteException {
	}

	/**
	 * Called when table record is about to be modified.
	 * At this point old values can still be obtained using
	 * <code>record.getValue()</code> call.
	 *
	 * @param record Record that is about to be modified.
	 * @param newValues New record values.
	 * @return possibly modified new values.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public Map recordModifying(int operationInitiator, SWRecord record,
							   Map newValues) throws SWFunctionalityException, RemoteException {
		if (AUTO_CHECK_PERMISSIONS)
			if (!isRecordModificationAllowed(operationInitiator, record, newValues))
				throw new SWFunctionalityException("Record modification as not allowed");

		return newValues;
	}

	/**
	 * Called when table record was modified.
	 * At this point <code>record</code> already contains new values and can't be modified.
	 *
	 * @param record Record that was modified.
	 * @param oldValues Old record values.
	 * @throws SWFunctionalityException If something
	 * was wrong during event handling.
	 *
	 * @see com.supportwizard.dml.ejb.SWRecordBean
	 */
	public void recordModified(int operationInitiator, SWRecord record,
							   Map oldValues) throws SWFunctionalityException, RemoteException {

		// do nothing
	}

	/**
	 * Called when the record is about to be removed.
	 *
	 * @param record Record that is about to be removed.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void recordRemoving(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException {
		if (AUTO_CHECK_PERMISSIONS)
			if (!isRecordRemovalAllowed(operationInitiator, record))
				throw new SWFunctionalityException("Record removal as not allowed");
	}

	/**
	 * Called when the record has been removed.
	 *
	 * @param operationInitiator A new operationInitiator for the operation, identified by it's code,
	 * see {@link SWFunctionalitiesConstants}
	 * @param recordData data from the removed record.
	 * @throws SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void recordRemoved(int operationInitiator, Map recordData) throws SWFunctionalityException, RemoteException {
	}

	/**
	 * Called when the table is about to be renamed.
	 * At this point old table name can still be obtained using
	 * <code>table.getName()</code> call (Each functinality must
	 * know about the table it is attached to).
	 *
	 * @param newName New table name.
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public void tableRenaming(String newName) throws SWFunctionalityException, RemoteException {
		if (AUTO_CHECK_PERMISSIONS)
			if (!isTableRenamingAllowed(newName))
				throw new SWFunctionalityException("Table renaming is not allowed");
	}

	// modification checking callbacks////////////////////////////////////////////////////

	/**
	 * Checks wether column with specified value object may be added.
	 *
	 * @return true if <code>column</code> may be added from <code>table</code>; false otherwise.
	 *
	 * @see com.supportwizard.dictionary.SWColumnBean
	 */
	public boolean isColumnAdditionAllowed(SWColumnValue columnValue) throws RemoteException {
		return true;
	}

	/**
	 * Checks wether specified column may be removed.
	 *
	 * @return true if <code>column</code> may be removed to <code>table</code>; false otherwise.
	 *
	 * @see com.supportwizard.dictionary.SWColumnBean
	 */
	public boolean isColumnRemovalAllowed(SWColumn column) throws RemoteException {
		return true;
	}

	/**
	 * Checks wether specified records may be imported.
	 *
	 * @param data data that is about to be imported.
	 *
	 * @return true if all of <code>data</code> rows are allowed to be imported.
	 */
	public boolean isRecordsImportAllowed(List data) throws RemoteException, SWFunctionalityException {
		for (int i = 0, datasize = data.size(); i < datasize; i++)
			if (!isRecordAdditionAllowed(SWFunctionalitiesConstants.BY_API, (Map) data.get(i)))
				return false;

		return true;
	}

	/**
	 * Checks wether specified record may be added.
	 *
	 * @return true if record containing <code>recordData</code> may be added; false otherwise.
	 *
	 * @see com.supportwizard.dml.ejb.SWRecordBean
	 */
	public boolean isRecordAdditionAllowed(int operationInitiator, Map recordData)
			throws RemoteException, SWFunctionalityException {
		return true;
	}

	/**
	 * Checks wether specified record may be removed.
	 *
	 * @return true if <code>record</code> may be removed; false otherwise.
	 *
	 * @see com.supportwizard.dml.ejb.SWRecordBean
	 */
	public boolean isRecordRemovalAllowed(int operationInitiator, SWRecord record)
			throws RemoteException, SWFunctionalityException {
		return true;
	}

	/**
	 * Checks wether specified record may be modified.
	 *
	 * @return true if <code>record</code> may be modified; false otherwise.
	 *
	 * @see com.supportwizard.dml.ejb.SWRecordBean
	 */
	public boolean isRecordModificationAllowed(int operationInitiator, SWRecord record, Map newValues)
			throws RemoteException, SWFunctionalityException {
		return true;
	}

	/**
	 * Checks wether specified table may be renamed.
	 *
	 * @return true if table may be renamed to <code>newName</code>; false otherwise.
	 */
	public boolean isTableRenamingAllowed(String newName) throws RemoteException {
		return true;
	}

	/**
	 * Called on reading read-only data. Usually for displaying the whole table
	 * in red-only mode. Thus it is unnecessary to fill in any additional SWData's
	 * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
	 * However, functionality logic may modify some fields or even add virtual fields.
	 *
	 * @param data Data being read. All functionalities will in series receive the same SWData instance.
	 *
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public SWData roDataReading(SWData data) throws SWFunctionalityException, RemoteException {
		return data;
	}

	/**
	 * Called on reading some database data that may be modified by user. Usually for displaying
	 * and modifying single line from the table. At this point functionality logic can insert some additional
	 * info into SWData's
	 * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
	 * Functionality logic may aslo modify some fields or add virtual fields.
	 *
	 * @param data Data being read. All functionalities will in series receive the same SWData instance.
	 *
	 * @throws com.supportwizard.functionalities.SWFunctionalityException If something
	 * was wrong during event handling.
	 */
	public SWData rwDataReading(SWData data) throws SWFunctionalityException, RemoteException {
		return data;
	}

	/**
	 * This callback should eliminate condition items based on artificial columns.
	 * It is impossible to directly query SW tables using artificial fiedls in WHERE clause;
	 * Functionality should implement this method, if it maintains some artificial fields and
	 * there's reasonable chance that client will use this field in ConditionList.
	 * <p>
	 * There are three options on how to maintain artificial fields in ConditionList:
	 * <ul>
	 * <li>Functionality can throw <code>SWFunctionalityException</code> if it does not want anybody
	 * to filter by particular artificial fiedl. It is acceptable, when filtering by some artificial field does not
	 * make sence or if it is very difficult to implement (filtering by history columns seems a good example here).
	 * Another example you can find in implementation of <tt>SWContactFunctionalityLogic</tt>, where functionality
	 * throws an exception when detects a constraint based on <tt>password</tt> field.
	 * </li>
	 * <li>Functionality can transform query parameters in such a way that they will not contain
	 * artificial constraints; it can substitute artificial filters with filters by non-artificial
	 * fields (see <tt>SWContacsFunctionalityLogic</tt> as possible example, where filtering by <tt>
	 * login</tt> field is substituted by filtering by main table's primary key). <b>Exercise caution
	 * when implementing this approach in a callback</b>. Make sure that <tt>ConditionList</tt> is still
	 * in consistent state after your modifications.
	 * </li>
	 * </ul>
	 * It is not recommended to simply remove <tt>ConditionElement</tt>s from <tt>ConditionList</tt>,
	 * without providing of an appropriate substitution with physical column filter.
	 * @param table Table we're working with.
	 * @param filters Query filters in the form of <tt>ConditionList</tt>. Functionality can modify this list.
	 * @throws SWFunctionalityException If evil comes
	 */
	public void prepareVirtualConstraints(SWTable table, ConditionList filters) throws RemoteException, SWFunctionalityException {

	}

	/**
	 * Gets an array of names of tables that must be dropped when functionality is removed.
	 * Here we mean not <code>SWTable</code> but simple database tables. In case if functionality
	 * wants to kill <code>SWTable</code> it must invoke <code>SWTable.remove()</code> itself
	 * (for example it it's own <code>ejbRemove()</code>) and return table's name here.
	 *
	 * This method helps to create more or less safe way to remove functionalities. The problem is in
	 * transactions handling: mySQL (and most other DBMS) can't rollback DDL commands (create table, drop
	 * table, alter table). Thus there can occur a situation when dictionary removes a number of fuctionalities
	 * associated with some table before removing the table itself and these functionalities wants to drop
	 * other tables. If funtionalities will invoke <code>SWInternalSQL.qlDDL(...)</code> themselves and
	 * some functionality will encounter allpication-level exception then the overall transaction
	 * cant' be correctly rolled back and this will lead to DB inconsistency.
	 *
	 * The solution is to execute all DDL instructions at the end of transaction. In this case all aplliction
	 * level exceptions can be safely processed and transaction can be rolled back.
	 *
	 * @return Array of names of tables that must be dropped when functionality is removed.
	 */
	public String[] getTablesToDropOnRemove() throws RemoteException {
		return null;
	}

	/**
	 * Sets functionality logic bus. Futher it can be accessed using private <code>bus</code>
	 * variable.
	 *
	 * @param bus Functionality logic bus.
	 */
	public void setBus(SWFunctionalityLogicBus bus) throws RemoteException {
		this.bus = bus;
	}

	/**
	 * Utility method for getting logger. Uses lazy initialization.
	 * It is really needed because this class is serializable and <code>Category</code>
	 * instance is declared as transient; thus it can be nullified at any moment.
	 *
	 * @return Logging category.
	 */
	protected Logger getLog() {
		if (log == null)
			log = Logger.getLogger(getClass());

		return log;
	}

	/**
	 * Gets <code>SimpleCache</code> instance.
	 *
	 * @return <code>SimpleCache</code> instance.
	 */
	protected SimpleCache getCache() {
		if (cache != null)
			return cache;

		cache = SimpleCacheFactory.getSimpleCache();

		return cache;
	}
}