/*
 * SupportWizard 2
 * 
 * Copyright (C) 2002 ISC corp. All Rights Reserved.
 * 
 * $Id: SWFunctionalityLogicBus.java,v 1.4 2003/07/22 23:05:30 alexi Exp $
 * Created by Konstantin Sobolev (kos@supportwizard.com)
 * Last modification $Date: 2003/07/22 23:05:30 $
 */
package com.supportwizard.functionalities;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Set;

/**
 * This bus is used to pass data between functionalities in the scope of
 * one callback. For each involved functionality broadcaster will call
 * {@link SWFunctionalityLogic#setBus(com.supportwizard.functionalities.SWFunctionalityLogicBus)}
 * method using the same bus instance. So functionalities can communicate, but only in one direction.<p>
 * For the pair callbacks like <code>recordAdding</code>--<code>recordAdded</code> functionality bus
 * is saved under the <code>"saved_bus"</code> key, so  you
 * can use<br>
 * <pre>
 * SWFunctionalityLogicBus savedBus = (SWFunctionalityLogicBus)currentBus.get("saved_bus");
 * </pre>
 * inside the <code>recordAdded</code> to get the bus as it existed at the end of
 * <code>recordAdding</code> procession.
 * 
 * @author <a href="mailto:kos@supportwizard.com" title="">Konstantin Sobolev</a>
 * @version $Revision: 1.4 $
 */
public class SWFunctionalityLogicBus implements Serializable {
  private HashMap storage = new HashMap();

  /**
   * Gets the number of elements in the bus.
   * 
   * @return the number of elements in the bus.
   */
  public int size() {
    return storage.size();
  } 

  /**
   * Returns <code>true</code> if this bus contains an object for the specified
   * key.
   * 
   * @return <code>true</code> if this map contains an object for the specified
   * key.
   * @param key key whose presence in this bus is to be tested.
   */
  public boolean containsKey(Object key) {
    return storage.containsKey(key);
  } 

  /**
   * Returns the value to which this bus maps the specified key.  Returns
   * <code>null</code> if the bus contains no mapping for this key.  A return
   * value of <code>null</code> does not <i>necessarily</i> indicate that the
   * bus contains no mapping for the key; it's also possible that the bus
   * explicitly maps the key to <code>null</code>.  The <code>containsKey</code>
   * operation may be used to distinguish these two cases.
   * 
   * @return the value to which this bus maps the specified key.
   * @param key key whose associated value is to be returned
   */
  public Object get(Object key) {
    return storage.get(key);
  } 

  /**
   * Associates the specified value with the specified key in this bus.
   * If the bus previously contained a mapping for this key, the old
   * value is replaced.
   * 
   * @param key key with which the specified value is to be associated.
   * @param value value to be associated with the specified key.
   * @return previous value associated with specified key, or <code>null</code>
   * if there was no mapping for key.  A <code>null</code> return can
   * also indicate that the bus previously associated
   * <code>null</code> with the specified key.
   */
  public Object put(Object key, Object value) {
    return storage.put(key, value);
  } 

  /**
   * Removes all elements from this bus.
   */
  public void clear() {
    storage.clear();
  } 

  /**
   * Returns a set view of the keys contained in this bus.  The set is
   * backed by the bus, so changes to the map are reflected in the bus, and
   * vice-versa.  The set supports element removal, which removes the
   * corresponding mapping from this bus, via the <code>Iterator.remove</code>,
   * <code>Set.remove</code>, <code>removeAll</code>, <code>retainAll</code>, and
   * <code>clear</code> operations.  It does not support the <code>add</code> or
   * <code>addAll</code> operations.
   * 
   * @return a set view of the keys contained in this map.
   */
  public Set keySet() {
    return storage.keySet();
  } 

  /**
   * Returns <code>true</code> if this bus contains no key-value mappings.
   * 
   * @return <code>true</code> if this bus contains no key-value mappings.
   */
  public boolean isEmpty() {
    return storage.isEmpty();
  } 

}

