/**
 * SupportWizard 2
 * Copyright (C) 2002 ISC corp. All Rights Reserved.
 * $Id: SWAccessFunctionalityLogic.java,v 1.7 2003/07/22 23:05:31 alexi Exp $
 */
package com.supportwizard.functionalities.access;

import com.supportwizard.cache.SimpleCache;
import com.supportwizard.cache.SimpleCacheFactory;
import com.supportwizard.cachewrappers.SWTableRelatedCachesWrapper;
import com.supportwizard.dictionary.SWChoiceLine;
import com.supportwizard.dictionary.SWColumn;
import com.supportwizard.dictionary.SWColumnHome;
import com.supportwizard.dictionary.SWColumnPK;
import com.supportwizard.dictionary.SWColumnValue;
import com.supportwizard.dictionary.SWDomainValue;
import com.supportwizard.dml.SWData;
import com.supportwizard.dml.interfaces.SWRecord;
import com.supportwizard.functionalities.SWFunctionalityException;
import com.supportwizard.functionalities.SWFunctionalityLogicBase;
import com.supportwizard.functionalities.interfaces.SWFuncColsMapping;
import com.supportwizard.functionalities.interfaces.SWFuncColsMappingHome;
import com.supportwizard.permissions.SWPermission;
import com.supportwizard.permissions.SWSubtypeColumnPermission;
import com.supportwizard.permissions.SWSubtypePermission;
import com.supportwizard.seance.Seance;
import com.supportwizard.subtypes.interfaces.SWSubtypeColumn;
import com.supportwizard.subtypes.interfaces.SWSubtypeColumnHome;
import com.supportwizard.subtypes.interfaces.SWSubtypeColumnPK;
import com.supportwizard.subtypes.ejb.SWSubtypeColumnValue;
import com.supportwizard.swgroups.SWGroupsPK;
import com.supportwizard.swusers.SWUsers;
import com.supportwizard.swusers.SWUsersHome;
import com.supportwizard.swusers.SWUsersPK;
import org.apache.log4j.Category;
import javax.ejb.FinderException;
import javax.naming.NamingException;
import java.rmi.RemoteException;
import java.security.Permissions;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * This class provides a set of methods to check client's permissions for record creation,
 * removal and modification. The client (and its permissions) are identified on the base of received <code>Seance</code>
 * instance.
 * <p>
 * This class is used by <strong>DML API</strong> facades - <code>SWDataModifier</code> and <code>SWRecordFacade</code>
 * to prevent unauthorised access to DML operations.
 * @see com.supportwizard.seance.Seance
 * @see com.supportwizard.dml.interfaces.SWDataModifier
 */
public class SWAccessFunctionalityLogic extends SWFunctionalityLogicBase {
  private final static Set specialDomains = new HashSet();
  private final static Set specialColumns = new HashSet();
  static {
    specialColumns.add("id");
    specialColumns.add("type");
    specialColumns.add("owner");
    specialColumns.add("workflow_state");
    specialColumns.add("_link_art_column");
    specialDomains.add("swdao3link");
  } 

  /**
   * Constructor declaration
   * 
   * 
   * @param funcID
   * @param tableID
   * 
   * @see
   */
  public SWAccessFunctionalityLogic(Long funcID, Long tableID) {
    super(funcID, tableID);
  }

  /**
   * Method declaration
   * 
   * 
   * @param operationInitiator
   * @param record
   * 
   * @return
   * 
   * @throws SWFunctionalityException
   * 
   * @see
   */
  public boolean isRecordRemovalAllowed(int operationInitiator, SWRecord record) throws SWFunctionalityException {
    try {
      Map values = record.getFieldValues();

      checkInitSeance(seance);

      Long ownerID = getOwnerID(seance, values, tableID);
      Long subtypeID = getSubtypeID(seance, values, tableID);

      if (subtypeID 
              == null) {  // Either we don't have subtype in a record, or we do not have mapping for subtype functionality. In both cases security is off.
        return true;
      } 

      SWSubtypePermission permRemoveOwn_SUB = new SWSubtypePermission(subtypeID, "delete_own");
      SWSubtypePermission permRemoveOth_SUB = new SWSubtypePermission(subtypeID, "delete_others");
      boolean             canRemove_SUB = checkPermissions(seance, permRemoveOwn_SUB, permRemoveOth_SUB, ownerID);

      if (canRemove_SUB 
              == false) {  // throw new DMLSecurityException(seance, "User do not have privilleges to remove records. UserID:" + seance.getUserID() + " SubtypeID:" + subtypeID);
        return false;
      } 

      return true;
    } catch (Exception e) {
      cat.error("Error occured while checking privilleges for records' removal", e);

      throw new SWFunctionalityException(e.getMessage());
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param operationInitiator
   * @param record
   * @param newValues
   * 
   * @return
   * 
   * @throws RemoteException
   * @throws SWFunctionalityException
   * 
   * @see
   */
  public boolean isRecordModificationAllowed(int operationInitiator, SWRecord record, 
                                             Map newValues) throws RemoteException, SWFunctionalityException {
    try {
      if (cat.isDebugEnabled()) {
        cat.debug("New values:" + newValues);
      } 

      Map oldValues = record.getFieldValues();

      checkInitSeance(seance);

      Long                ownerID = getOwnerID(seance, oldValues, tableID);
      Long                subtypeID = getSubtypeID(seance, oldValues, tableID);
      SWSubtypePermission permModifOwn_SUB = new SWSubtypePermission(subtypeID, "edit_own");
      SWSubtypePermission permModifOth_SUB = new SWSubtypePermission(subtypeID, "edit_others");
      boolean             canChange_SUB = checkPermissions(seance, permModifOwn_SUB, permModifOth_SUB, ownerID);

      if (canChange_SUB 
              == false) {  // throw new DMLSecurityException(seance, "User do not have privilleges to modify the record. UserID:" + seance.getUserID() + " SubtypeID:" + subtypeID);
        return false;
      } 

      SWTableRelatedCachesWrapper cachesWrapper = new SWTableRelatedCachesWrapper();
      Map                         names2DomainValues = cachesWrapper.getNames2DomainValues(seance, tableID);
      Set                         newValuesEntries = newValues.entrySet();

      for (Iterator it = newValuesEntries.iterator(); it.hasNext(); ) {
        Map.Entry     fieldEntry = (Map.Entry) it.next();
        String        fieldName = (String) fieldEntry.getKey();
        SWDomainValue domainValue = (SWDomainValue) names2DomainValues.get(fieldName);
        String        domainName = domainValue.getName();
        boolean       skip = isSpecialColumn(fieldName, domainName);

        if (skip) {
          continue;
        } 

        Object                    fieldValue_NEW = fieldEntry.getValue();
        SWSubtypeColumnPK         subtypeColumnPK = getCachedSWSubtypeColumnPK(seance, tableID, fieldName, 
                subtypeID.longValue());
        SWSubtypeColumnPermission permModifOwn_FLD = new SWSubtypeColumnPermission(subtypeColumnPK, "edit_own");
        SWSubtypeColumnPermission permModifOth_FLD = new SWSubtypeColumnPermission(subtypeColumnPK, "edit_others");
        boolean                   canChange_FLD = checkPermissions(seance, permModifOwn_FLD, permModifOth_FLD, ownerID);

        if (canChange_FLD == true) {
          if (cat.isDebugEnabled()) {
            cat.debug("Can edit:" + canChange_FLD);
          } 

          continue;
        } 

        Object fieldValue_OLD = oldValues.get(fieldName);

        if (cat.isDebugEnabled()) {
          cat.debug("Field:" + fieldName + " New Value:\"" + fieldValue_NEW + "\" Old Value: \"" + fieldValue_OLD + "\"");
        } 

        Object dummy = new Object();

        if (fieldValue_NEW == null) {
          fieldValue_NEW = dummy;
        } 

        if (fieldValue_OLD == null) {
          fieldValue_OLD = dummy;
        } 

        if (fieldValue_NEW.equals(fieldValue_OLD)) {
          cat.debug("Column value passed for record modification is equal to its previous value. Good luck...");

          continue;
        }  // throw new DMLSecurityException(seance, "User do not have privilleges to modify the field of the record. UserID:" + seance.getUserID() + " SubtypeID:" + subtypeID + " field name:" + fieldName);
 
        return false;
      } 

      return true;
    } catch (Exception e) {
      cat.error("Error occured while checking privilleges for records' modification", e);

      throw new SWFunctionalityException(e.getMessage());
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param operationInitiator
   * @param values
   * 
   * @return
   * 
   * @throws RemoteException
   * @throws SWFunctionalityException
   * 
   * @see
   */
  public boolean isRecordAdditionAllowed(int operationInitiator, 
                                         Map values) throws RemoteException, SWFunctionalityException {
    try {
      checkInitSeance(seance);

      Long subtypeID = getSubtypeID(seance, values, tableID);

      if (subtypeID 
              == null) {  // Either we don't have subtype in a record, or we do not have mapping for subtype functionality. In both cases security is off.
        return true;
      } 

      SWSubtypePermission permCreate_SUB = new SWSubtypePermission(subtypeID, "create");
      boolean             canCreate_SUB = checkPermissions(seance, permCreate_SUB, null, null);

      if (canCreate_SUB 
              == false) {  // throw new DMLSecurityException(seance, "User do not have privilleges to create the record. UserID:" + seance.getUserID() + " SubtypeID:" + subtypeID);
        return false;
      } 

      SWTableRelatedCachesWrapper cachesWrapper = new SWTableRelatedCachesWrapper();
      Map                         names2DomainValues = cachesWrapper.getNames2DomainValues(seance, tableID);
      Set                         valuesEntries = values.entrySet();

      for (Iterator it = valuesEntries.iterator(); it.hasNext(); ) {
        Map.Entry     fieldEntry = (Map.Entry) it.next();
        String        columnName = (String) fieldEntry.getKey();
        SWDomainValue domainValue = (SWDomainValue) names2DomainValues.get(columnName);
        String        domainName = domainValue.getName();
        boolean       skip = isSpecialColumn(columnName, domainName);

        if (skip) {
          continue;
        } 

        Object               fieldValue = fieldEntry.getValue();
        SWSubtypeColumnPK    subtypeColumnPK = getCachedSWSubtypeColumnPK(seance, tableID, columnName, 
                subtypeID.longValue());
        SWSubtypeColumnValue subtypeColumnValue = getCachedSWSubtypeColumnValue(seance, subtypeColumnPK);

        if (subtypeColumnValue.getSwTableID() == -1 && subtypeColumnValue.getSwSubtypeID() == -1) {
          continue;
        } 

        SWSubtypeColumnPermission permCreate_FLD = new SWSubtypeColumnPermission(subtypeColumnPK, "create");
        boolean                   canCreate_FLD = checkPermissions(seance, permCreate_FLD, null, null);

        if (canCreate_FLD == true) {
          cat.debug("User has privilleges to create the column:" + columnName);

          continue;
        } 

        cat.info("User has no privilleges to create the column:" + columnName);

        String subtypeDefValue = subtypeColumnValue.getDefaultValue();
        String dummy = "";

        if (subtypeDefValue == null) {
          subtypeDefValue = dummy;
        } 

        if (fieldValue == null) {
          fieldValue = dummy;
        } 

        final String fieldValueStr = fieldValue.toString().trim();
        final String domainDefValueTrim = subtypeDefValue.trim();

        if (domainDefValueTrim.equals(fieldValueStr)) {
          cat.debug("Column value passed for record creation is equal to its subtype default value. Good luck...");

          continue;
        }  // throw new DMLSecurityException(seance, "User do not have privilleges to create the field of the record. UserID:" + seance.getUserID() + " SubtypeID:" + subtypeID + " Field Name:" + columnName + " Field Value:\"" + fieldValueStr + "\" Domain Default Value:\"" + domainDefValueTrim + "\"");
 
        return false;
      } 

      return true;
    } catch (Exception e) {
      cat.error("Error occured while checking privilleges for records' creation");

      throw new SWFunctionalityException(e.getMessage());
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param data
   * 
   * @return
   * 
   * @throws RemoteException
   * @throws SWFunctionalityException
   * 
   * @see
   */
  public SWData roDataReading(SWData data) throws SWFunctionalityException, RemoteException {
    return dataReading(seance, data);
  } 

  /**
   * Called on reading some database data that may be modified by user. Usually for displaying
         * and modifying single line from the table. At this point functionality logic can insert some additional
         * info into SWData's
         * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
         * Functionality logic may aslo modify some fields or add virtual fields.
         *
         * @param data Data being read. All functionalities will in series receive the same SWData instance.
         *
         * @throws com.supportwizard.functionalities.SWFunctionalityException If something
         * was wrong during event handling.
   */
  public SWData rwDataReading(SWData data) throws SWFunctionalityException, RemoteException {
    return dataReading(seance, data);
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param data
   * 
   * @return
   * 
   * @throws SWFunctionalityException
   * 
   * @see
   */
  static private SWData dataReading(Seance seance, SWData data) throws SWFunctionalityException {
    try {
      checkInitSeance(seance);

      Long                        tableID = data.getTableID();
      SWTableRelatedCachesWrapper cachesWrapper = new SWTableRelatedCachesWrapper();
      Map                         names2DomainValues = cachesWrapper.getNames2DomainValues(seance, tableID);
      SimpleCache                 cache = SimpleCacheFactory.getSimpleCache();
      int                         dataSize = data.size();

      for (int dataIdx = 0; dataIdx < dataSize; dataIdx++) {
        Map  values = data.getRowAsMap(dataIdx);
        Long ownerID = getOwnerID(seance, values, tableID);
        Long subtypeID = getSubtypeID(seance, values, tableID);

        if (cat.isDebugEnabled()) {
          cat.debug("Next SWData record:" + values + "Subtype:" + subtypeID + " Owner:" + ownerID);
        } 

        if (subtypeID 
                == null) {  // Either we don't have subtype in a record, or we do not have mapping for subtype functionality. In both cases security is off.
          continue;
        } 

        Set    fieldsToNullify = null;
        String cacheKey = "subtype." + subtypeID + ".nullify";

        if ((fieldsToNullify = (Set) cache.get(cacheKey)) == null) {
          if (cat.isInfoEnabled()) {
            cat.info("Fields to nullify not in cache:" + cacheKey);
          } 

          fieldsToNullify = new HashSet();

          Set  valuesEntries = values.entrySet();
          Long userID = seance.getUserID();

          if (cat.isDebugEnabled()) {
            cat.debug("Owner ID:" + ownerID + "User ID:" + userID);
          } 

          for (Iterator it = valuesEntries.iterator(); it.hasNext(); ) {
            Map.Entry     fieldEntry = (Map.Entry) it.next();
            String        fieldName = (String) fieldEntry.getKey();
            SWDomainValue domainValue = (SWDomainValue) names2DomainValues.get(fieldName);
            String        domainName = domainValue.getName();
            boolean       skip = isSpecialColumn(fieldName, domainName);

            if (skip) {
              continue;
            } 

            SWSubtypeColumnPK         subtypeColumnPK = getCachedSWSubtypeColumnPK(seance, tableID, fieldName, 
                    subtypeID.longValue());
            SWSubtypeColumnPermission permViewOwn_FLD = new SWSubtypeColumnPermission(subtypeColumnPK, "view_own");
            SWSubtypeColumnPermission permViewOth_FLD = new SWSubtypeColumnPermission(subtypeColumnPK, "view_others");
            boolean                   canView_FLD = checkPermissions(seance, permViewOwn_FLD, permViewOth_FLD, ownerID);

            if (canView_FLD == false) {
              cat.info("User do not have privilleges to view the field of the record. Restricted field:" + fieldName);
              fieldsToNullify.add(fieldName);
            } 
          } 

          cache.put(cacheKey, fieldsToNullify);
        } 

        for (Iterator it = fieldsToNullify.iterator(); it.hasNext(); ) {
          String fieldName = (String) it.next();

          values.put(fieldName, null);
        } 

        data.setRowAsMap(dataIdx, values);
      } 

      return data;
    } catch (Exception e) {
      cat.error("Error occured while checking privilleges for data reading");

      throw new SWFunctionalityException(e.getMessage());
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param permOwn
   * @param permOthr
   * @param owner
   * 
   * @return
   * 
   * @throws FinderException
   * @throws NamingException
   * @throws RemoteException
   * 
   * @see
   */
  static boolean checkPermissions(Seance seance, SWPermission permOwn, SWPermission permOthr, 
                                  Long owner) throws RemoteException, FinderException, NamingException {
    Long dummy = new Long(-999);
    Long user = seance.getUserID();

    if (user == null) {
      user = dummy;
    } 

    if (owner == null) {
      owner = user;  // Treat all the records as someones' own, if no owner information detected.
    } 

    Permissions permissions = getCachedPermissions(seance);

    if (cat.isDebugEnabled()) {
      cat.debug("Checking element. Owner ID:" + owner + " user ID:" + user + " Permission own:" + permOwn + "perm others:" 
                + permOthr);
    } 

    if (owner.equals(user)) {  // Check element as our own.
      cat.debug("Check own permissions...");

      if (permissions.implies(permOwn)) {
        return true;
      } else {
        return false;
      } 
    } else {                   // Check element as others'
      cat.debug("Check others permissions...");

      if (permissions.implies(permOthr)) {
        return true;
      } else {
        return false;
      } 
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param values
   * @param tableID
   * @param alias
   * 
   * @return
   * 
   * @throws FinderException
   * @throws IllegalAccessException
   * @throws NamingException
   * @throws NoSuchFieldException
   * @throws RemoteException
   * 
   * @see
   */
  static private Object getCachedValueByAlias(Seance seance, Map values, Long tableID, String alias) 
          throws NamingException, NoSuchFieldException, IllegalAccessException, FinderException, RemoteException {
    cat.debug("ENTERING getValueByFuncColsMappingAlias: tableID:" + tableID + " alias:" + alias);

    String                      cacheKey = "col_alias." + alias;
    SWTableRelatedCachesWrapper cachesWrapper = new SWTableRelatedCachesWrapper();
    Object                      valueOfAlias = null;

    if ((valueOfAlias = cachesWrapper.getUserObject(seance, tableID, cacheKey)) == null) {
      if (cat.isInfoEnabled()) {
        cat.info("Value Of Alias Not in cache:" + cacheKey);
      } 

      SWFuncColsMappingHome funcColsMappingHome = (SWFuncColsMappingHome) seance.getHome(SWFuncColsMappingHome.class);
      Collection            mappings = null;

      try {
        mappings = funcColsMappingHome.findByCombination(null, tableID, null, alias);
      } catch (FinderException e) {
        mappings = new ArrayList();
      } 

      if (cat.isDebugEnabled()) {
        cat.debug("Collection of mappings obtained. Its size:" + mappings.size());
      } 

      if (mappings.size() > 1) {
        throw new IllegalStateException("Wrong number of column mappings for the given alias - should be at most 1. Alias:" 
                                        + alias);
      } 

      if (mappings.size() == 0) {  // Null return will signalize us that we do not have a mapping requested.
        return null;
      } 

      SWFuncColsMapping mapping = (SWFuncColsMapping) mappings.iterator().next();
      SWColumn          column = mapping.getColumn();
      String            fieldName = column.getName();

      cat.debug("Mapped column name obtained:" + fieldName);

      valueOfAlias = values.get(fieldName);

      cachesWrapper.putUserObject(seance, tableID, cacheKey, valueOfAlias);
    } 

    return valueOfAlias;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param values
   * @param tableID
   * 
   * @return
   * 
   * @throws FinderException
   * @throws IllegalAccessException
   * @throws NamingException
   * @throws NoSuchFieldException
   * @throws RemoteException
   * 
   * @see
   */
  static private Long getOwnerID(Seance seance, Map values, Long tableID) 
          throws NoSuchFieldException, NamingException, RemoteException, FinderException, IllegalAccessException {
    Object ownerObj = 
      getCachedValueByAlias(seance, values, tableID, 
                            "owner_column_alias");  /* Essential hack - fulfillment of Gof's request from  July, 2, 2002.
    There is no guarantee that map will contain ownerID value in the form of Long (it can be String) */
    String dummy = "";

    if (ownerObj == null) {
      ownerObj = dummy;
    } 

    Long ownerID;

    if (ownerObj.equals(dummy)) {
      cat.warn("Record passed does not contain owner field value.");

      ownerID = null;
    } else {
      ownerID = (ownerObj instanceof SWChoiceLine) ? ((SWChoiceLine) ownerObj).getId() : new Long(ownerObj.toString());
    } 

    return ownerID;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param values
   * @param tableID
   * 
   * @return
   * 
   * @throws FinderException
   * @throws IllegalAccessException
   * @throws NamingException
   * @throws NoSuchFieldException
   * @throws RemoteException
   * 
   * @see
   */
  static private Long getSubtypeID(Seance seance, Map values, Long tableID) 
          throws NoSuchFieldException, FinderException, IllegalAccessException, RemoteException, NamingException {
    Object subtypeObj = getCachedValueByAlias(seance, values, tableID, "subtype_column_alias");

    if (subtypeObj 
            == null) {  // Either we don't have subtype in a record, or we do not have mapping for subtype functionality. In both cases security is off.
      cat.warn("subtype is null ! ");

      return null;
    } 

    Long subtypeID = new Long(subtypeObj.toString());

    return subtypeID;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param tableID
   * @param fieldName
   * @param subtypeID
   * 
   * @return
   * 
   * @throws FinderException
   * @throws NamingException
   * @throws RemoteException
   * 
   * @see
   */
  static private SWSubtypeColumnPK getCachedSWSubtypeColumnPK(Seance seance, Long tableID, String fieldName, 
          long subtypeID) throws NamingException, FinderException, RemoteException {
    String                      cacheKey = "column_pk." + fieldName;
    SWTableRelatedCachesWrapper cachesWrapper = new SWTableRelatedCachesWrapper();
    SWColumnPK                  columnPK = null;

    if ((columnPK = (SWColumnPK) cachesWrapper.getUserObject(seance, tableID, cacheKey)) == null) {
      if (cat.isInfoEnabled()) {
        cat.info("Column PK is not in cache:" + cacheKey);
      } 

      SWColumnHome  columnHome = (SWColumnHome) seance.getHome(SWColumnHome.class, "dictionary/SWColumn");
      SWColumn      column = columnHome.findByName(tableID, fieldName);
      Long          languageID = seance.getLanguageID();
      SWColumnValue columnValue = column.getColumnValue(languageID);

      columnPK = columnValue.getSWColumnPK();

      cachesWrapper.putUserObject(seance, tableID, cacheKey, columnPK);
    } 

    long              columnID = columnPK.getSWColumnID().longValue();
    SWSubtypeColumnPK subtypeColumnPK = new SWSubtypeColumnPK(tableID, new Long(subtypeID), new Long(columnID));

    if (cat.isDebugEnabled()) {
      cat.debug("SWSubtypeColumnPK Created:" + subtypeColumnPK.toString());
    } 

    return subtypeColumnPK;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * @param subtypeColumnPK
   * 
   * @return
   * 
   * @throws NamingException
   * @throws RemoteException
   * 
   * @see
   */
  static private SWSubtypeColumnValue getCachedSWSubtypeColumnValue(Seance seance, 
          SWSubtypeColumnPK subtypeColumnPK) throws NamingException, RemoteException {
    String               cacheKey = "subtype." + subtypeColumnPK.getSwSubtypeID() + "." + subtypeColumnPK.getSwColumnID();
    SWSubtypeColumnValue subtypeColumnValue = null;
    SimpleCache          cache = SimpleCacheFactory.getSimpleCache();

    if ((subtypeColumnValue = (SWSubtypeColumnValue) cache.get(cacheKey)) == null) {
      if (cat.isInfoEnabled()) {
        cat.info("SWSubtypeColumnValue is not in cache:" + cacheKey);
      } 

      SWSubtypeColumnHome subtypeColumnHome = (SWSubtypeColumnHome) seance.getHome(SWSubtypeColumnHome.class, 
              SWSubtypeColumnHome.JNDI_NAME);

      try {
        SWSubtypeColumn subtypeColumn = subtypeColumnHome.findByPrimaryKey(subtypeColumnPK);
        Long            langueID = seance.getLanguageID();

        subtypeColumnValue = subtypeColumn.getSubtypeColumnValue(langueID);
      } catch (FinderException e) {
        cat.info("Subtype column for the following PK is not found:" + subtypeColumnPK);

        subtypeColumnValue = new SWSubtypeColumnValue();

        subtypeColumnValue.setSwSubtypeID(-1);
        subtypeColumnValue.setSwTableID(-1);
      } 

      cache.setExpirationPeriod(cacheKey, 300000L);  // 5min
      cache.put(cacheKey, subtypeColumnValue);
    } 

    return subtypeColumnValue;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * 
   * @return
   * 
   * @throws FinderException
   * @throws NamingException
   * @throws RemoteException
   * 
   * @see
   */
  static private Permissions getCachedPermissions(Seance seance) throws NamingException, FinderException, RemoteException {
    Long        userID = seance.getUserID();
    Permissions permissions = null;
    SimpleCache cache = SimpleCacheFactory.getSimpleCache();
    String      cacheKey = "user." + userID + ".permissions";

    if ((permissions = (Permissions) cache.get(cacheKey)) == null) {
      if (cat.isInfoEnabled()) {
        cat.info("User Permissions are not in cache:" + cacheKey);
      } 

      SWUsers user = (SWUsers) seance.getUser();

      permissions = user.getUserPermissions();

      cache.setExpirationPeriod(cacheKey, 300000L);  // 5min
      cache.put(cacheKey, permissions);
    } 

    return permissions;
  } 

  /**
   * Method declaration
   * 
   * 
   * @param seance
   * 
   * @see
   */
  static private void checkInitSeance(Seance seance) {
    if (seance == null) {
      cat.error("Seance has not been set.");

      throw new IllegalStateException("Seance has not been set.");
    } 
  } 

  /**
   * Method declaration
   * 
   * 
   * @param fieldName
   * @param domainName
   * 
   * @return
   * 
   * @see
   */
  static boolean isSpecialColumn(String fieldName, 
                                 String domainName) {  // NOT to check any "special" columns, specified by their names.
    if (specialColumns.contains(fieldName.toLowerCase())) {
      if (cat.isDebugEnabled()) {
        cat.debug("Excluding special column from permission checking because of its name. Column Name:" + fieldName);
      } 

      return true;
    } 

    if (cat.isDebugEnabled()) {
      cat.debug("Domain name:" + domainName);  // NOT to check any "special" columns, specified by their domains' names.
    } 

    if (specialDomains.contains(domainName.toLowerCase())) {
      if (cat.isDebugEnabled()) {
        cat.debug("Excluding special column from permissions checking because of domain name. Column Name:" + fieldName 
                  + " Domain Name:" + domainName);
      } 

      return true;
    } 

    return false;
  } 

  private static Category cat = Category.getInstance(SWAccessFunctionalityLogic.class.getName());
}

