/*
 * SupportWizard 2
 * 
 * Copyright (C) 2002 ISC corp. All Rights Reserved.
 * 
 * $Id: DbAHstTableManagerImpl.java,v 1.5 2003/07/22 23:05:31 alexi Exp $
 * Created by Konstantin Sobolev (kos@supportwizard.com) on Jun 23, 2002
 * Last modification $Date: 2003/07/22 23:05:31 $
 */
package com.supportwizard.functionalities.attributehistory;

import com.supportwizard.db.*;
import com.supportwizard.dictionary.SWColumn;
import com.supportwizard.dictionary.SWTable;
import com.supportwizard.dictionary.SWTypeValue;
import com.supportwizard.functionalities.attributehistory.ejb.SWAttributeHistoryLogRecordConst;
import com.supportwizard.dictionary.util.ColumnNamesHelper;
import com.supportwizard.utils.EJBException2;
import org.apache.log4j.Category;
import javax.ejb.EJBException;
import java.rmi.RemoteException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * Utility class for managing _ahst_ tables. Mainly used by <code>SWAttributeHistoryFunctionalityBean</code>.
 * <p>
 * _ahst_ table is used for storing <code>SWAttributeHistoryLogRecordConst</code> data and has the following layout:
 * <ul>
 * <li>BIGINT RecordID  -- Autoincrementing primary key.</li>
 * <li>BIGINT ActionDate -- Action date in milliseconds.</li>
 * <li>BIGINT UserID -- Actor's ID.</li>
 * <li>{set of fields that make up user table's primary key with '_pk_' suffixes in their names} -- used for
 * storing <code>SWRecordPK</code></li>
 * </ul>
 * 
 * @author <a href="mailto:kos@supportwizard.com" title="">Konstantin Sobolev</a>
 * @version $Revision: 1.5 $
 * @see SWAttributeHistoryLogRecordConst
 * @deprecated Use DictionaryAHstTableManagerImpl
 */
public class DbAHstTableManagerImpl implements AHstTableManager {
  private final Category cat = Category.getInstance(DbAHstTableManagerImpl.class);

  /**
   * Creates new _ahst_ table.
   * 
   * @param table Target user table.
   * 
   * @throws EJBException in case of run-time error.
   */
  public void createHstTable(SWTable table) {
    SWInternal_SQL sw_sql = null;

    try {
      sw_sql = getSQLManager();
    } catch (SW_SQLException e) {
      cat.error("Couldn't get DB connection", e);

      throw new EJBException(e);
    } 

    try {
      FieldList fieldList = new FieldList();

      fieldList.add("RecordID", Types.BIGINT, false, "", true, true);                                   // autoincrementing PK
      fieldList.add("UserID", Types.BIGINT, false, "", false, false);
      fieldList.add("ActionDate", Types.BIGINT, false, "", false, false);                               // 

      ArrayList columns = table.getSWColumns();                                                         // all table columns
      HashSet   pkColumnNames = new ColumnNamesHelper().getPrimaryKeyFieldNames(table.getSWTableID());  // columns forming pk

      for (Iterator it = columns.iterator(); it.hasNext(); ) {

        // iterate thru all table columns
        SWColumn    column = (SWColumn) it.next();
        String      columnName = column.getDBName();                                                    // we are using db names!!
        SWTypeValue typeValue = column.getSWDomain().getSWType().getTypeValue();

        if (pkColumnNames.contains(columnName)) {                                                       // if this column is included into pk
          fieldList.add(columnName + SWAttributeHistoryLogRecordConst.PK_FIELD_NAMES_SUFFIX, 
                        typeValue.getFullLocalDataType(), false, "", false, false);
        } 
      } 

      HashMap params = new HashMap();

      params.put("table", getAHstTableName(table));
      params.put("existing", "y");                                                                      // directs to use 'drop table if exists' instead of 'drop table'

      try {
        sw_sql.sqlDDL("drop table", params);                                                            // to be on the safe side
      } catch (Exception e) {                                                                           /* skip */
        cat.warn("Exception on drop", e);
      } 

      params.put("fields", fieldList);
      params.put("primary_name", "RecordID");
      sw_sql.sqlDDL("create table", params);
      cat.debug("Created " + getAHstTableName(table) + " table");

      // new TableDumper().dumpTableStructure(getAHstTableName());
    } catch (Throwable e) {
      cat.debug("Exception caught", e);

      throw new EJBException2(e);
    } 
    finally {
      try {
        sw_sql.dbDisconnect();
      } catch (SQLException e) {
        cat.error("Cant't close database connection", e);
      } 
    } 
  } 

  /**
   * Gets the number of records in an _ahst_ table.
   * 
   * @return Number of records in _ahst_ table.
   * 
   * @throws EJBException in case of run-time error.
   */
  public int getAHstTableSize(Long tableID) {
    int            res = -1;
    SWInternal_SQL sw_sql = null;

    try {
      sw_sql = getSQLManager();
    } catch (SW_SQLException e) {
      cat.error("Couldn't get DB connection", e);

      throw new EJBException(e);
    } 

    try {
      TableList tableList = new TableList();

      tableList.add(getAHstTableName(tableID), "");

      ColumnList columnList = new ColumnList();

      columnList.add("*", "", "", "count");

      HashMap params = new HashMap();

      params.put("from", tableList);  // select count(*) from blablabla_ahst_
      params.put("columns", columnList);
      params.put("where", new ConditionList());

      ResultSet rs = sw_sql.sqlQuery(params);

      rs.next();

      res = rs.getInt(1);

      rs.close();
    } catch (Exception e) {
      cat.debug("Exception caught", e);

      throw new EJBException2(e);
    } 
    finally {
      try {
        sw_sql.dbDisconnect();
      } catch (SQLException e) {
        cat.error("Cant't close database connection", e);
      } 
    } 

    return res;
  } 

  /**
   * Drops _ahst_ table.
   * 
   * @param tableID ID of user table _ahst_ table is created for.
   * 
   * @throws EJBException in case of run-time error.
   */
  public void dropAHstTable(Long tableID) {
    SWInternal_SQL sw_sql = null;

    try {
      sw_sql = getSQLManager();
    } catch (SW_SQLException e) {
      cat.error("Couldn't get DB connection", e);

      throw new EJBException(e);
    } 

    try {
      HashMap params = new HashMap();

      params.put("table", getAHstTableName(tableID));
      sw_sql.sqlDDL("drop table", params);
    } catch (Exception e) {
      cat.debug("Exception caught", e);

      throw new EJBException2(e);
    } 
    finally {
      try {
        sw_sql.dbDisconnect();
      } catch (SQLException e) {
        cat.error("Cant't close database connection", e);
      } 
    } 
  } 

  // ___________________________private area___________________________________________________________________________

  /**
   * Gets _ahst_ table name.
   * 
   * @param table Target user table.
   * @return _ahst_ table name.
   */
  public String getAHstTableName(SWTable table) {
    try {
      return table.getSWTableID().toString() + SWAttributeHistoryLogRecordConst.ATTR_HISTORY_TABLE_NAME_SUFFIX;
    } catch (RemoteException e) {
      cat.error("Communication Error: " + e, e);

      throw new EJBException(e);
    } 
  } 

  /**
   * Gets _ahst_ table name.
   * 
   * @param tableID ID of target user table.
   * @return _ahst_ table name.
   */
  public String getAHstTableName(Long tableID) {
    return tableID.toString() + SWAttributeHistoryLogRecordConst.ATTR_HISTORY_TABLE_NAME_SUFFIX;
  } 

  public String[] getTablesToDropOnRemove(Long tableID) {
    return new String[] {
      getAHstTableName(tableID)
    };
  } 

  private SWInternal_SQL getSQLManager() throws SW_SQLException {
    final SWInternal_SQL sw_sql;

    sw_sql = new SWInternal_SQL();

    sw_sql.dbConnect();

    return sw_sql;
  } 

}

