/*
 * SupportWizard 2
 * 
 * Copyright (C) 2002 ISC corp. All Rights Reserved.
 * 
 * $Id: SWAttributeHistoryFunctionalityLogic.java,v 1.13 2003/07/22 23:05:31 alexi Exp $
 * Created by Konstantin Sobolev (kos@supportwizard.com) on 27.06.2002
 * Last modification $Date: 2003/07/22 23:05:31 $
 */
package com.supportwizard.functionalities.attributehistory;

import com.supportwizard.functionalities.SWFunctionalityLogicBase;
import com.supportwizard.functionalities.SWFunctionalityException;
import com.supportwizard.functionalities.attributehistory.ejb.SWAttributeHistoryFunctionalityConst;
import com.supportwizard.functionalities.attributehistory.interfaces.*;
import com.supportwizard.dml.SWData;
import com.supportwizard.dml.SWRecordPK;
import com.supportwizard.dml.interfaces.SWRecord;
import com.supportwizard.utils.HomesGetter;
import com.supportwizard.utils.EJBException2;
import com.supportwizard.utils.StringParams;
import com.supportwizard.swgroups.SWGroupsPK;
import com.supportwizard.swusers.SWUsersPK;
import com.supportwizard.swusers.SWUsersHome;
import javax.naming.NamingException;
import javax.ejb.FinderException;
import java.rmi.RemoteException;
import java.util.*;

// ejb-refs to:
// SWAttributeHistoryLogRecord
// SWAttributeChange
// SWUsers

/**
 * Attribute history functionality logic. Processess the following events:
 * <table border="1">
 * <tr><td>Event</td><td>Action taken</td></tr>
 * <tr><td>recordRemoving</td><td>Removes all attribute history of the reccord being removed</td></tr>
 * <tr><td>roDataReading</td><td>Creates virtual column with name =
 * <code>{@link SWAttributeHistoryFunctionalityConst#HISTORY_FIELD_NAME HISTORY_FIELD_NAME}</code>
 * and for each read record writes there all attribute changes history in human-readable form</td></tr>
 * </table>
 * 
 * @author <a href="mailto:kos@supportwizard.com" title="">Konstantin Sobolev</a>
 * @version $Revision: 1.13 $
 */
public class SWAttributeHistoryFunctionalityLogic extends SWFunctionalityLogicBase {
  private SWAttributeHistoryLogRecordHome attributeHistoryLogRecordHome;
  private SWAttributeChangeHome           attributeChangeHome;
  private SWUsersHome                     usersHome;
  private final String                    lineSep = System.getProperty("line.separator");

  /**
   * Basic constructor. To be called by children.
   */
  public SWAttributeHistoryFunctionalityLogic(Long funcID, Long tableID) {
    super(funcID, tableID);
  }

  /**
   * Called when the record is about to be removed. Deletes all attribute history of this record.
   * 
   * @param record Record that is about to be removed.
   * @throws SWFunctionalityException If something
   * was wrong during event handling.
   */
  public void recordRemoving(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException {
    HomesGetter homesGetter = getSeance().getHomesGetter();

    try {
      getHomes(homesGetter);

      SWRecordPK recordPK = (SWRecordPK) record.getPrimaryKey();

      getLog().debug("Record with pk=" + recordPK + " is being removed; killing attribute history");

      // get all history log records for this record
      Collection history = attributeHistoryLogRecordHome.findBySWRecordPK(recordPK);

      if (getLog().isDebugEnabled()) {
        getLog().debug("Got " + history.size() + " history records");
      } 

      for (Iterator it = history.iterator(); it.hasNext(); ) {
        SWAttributeHistoryLogRecord   logRecord = (SWAttributeHistoryLogRecord) it.next();
        SWAttributeHistoryLogRecordPK pk = (SWAttributeHistoryLogRecordPK) logRecord.getPrimaryKey();
        Long                          recordID = pk.getRecordID();
        Collection                    changes = attributeChangeHome.findByRecordID(recordID);

        for (Iterator it2 = changes.iterator(); it2.hasNext(); ) {
          SWAttributeChange attributeChange = (SWAttributeChange) it2.next();

          attributeChange.remove();
        } 

        logRecord.remove();
      } 
    } catch (Throwable e) {
      getLog().error("Error removing undo data", e);

      throw new EJBException2(e, "Error removing undo data");
    } 
  } 

  /**
   * Called on reading read-only data. Creates virtual column with name =
   * <code>{@link SWAttributeHistoryFunctionalityConst#HISTORY_FIELD_NAME HISTORY_FIELD_NAME}</code>
   * and for each read record writes there all attribute changes history in human-readable form.
   * 
   * @param data Data being read.
   */
  public SWData roDataReading(SWData data) throws RemoteException {

    // first determine if client really needs 'Attribute History' column to be calculated
    List metadata = data.getSWMeta().getColumnList();

    if (!metadata.contains(SWAttributeHistoryFunctionalityConst.HISTORY_FIELD_NAME)) {
      return data;  // very effective optimization

      // it seems to me that this code is easier to maintain than classical lazy initialization like
      // AttributeHistoryLogRecordHome attributeHistoryLogRecordHome = getAttributeHistoryLogRecordHome();
    } 

    try {
      getHomes(getSeance().getHomesGetter());  // now we can safely use class variables
    } catch (Throwable e) {
      getLog().error("Couldn't get home interfaces", e);

      throw new EJBException2(e, "Couldn't get home interfaces");
    } 

    // --
    String columnName = getParameter(SWAttributeHistoryFunctionalityConst.HISTORY_FIELD_NAME_ATTR);

    if (columnName == null) {
      getLog().warn(SWAttributeHistoryFunctionalityConst.HISTORY_FIELD_NAME_ATTR + " parameter not found");

      columnName = SWAttributeHistoryFunctionalityConst.HISTORY_FIELD_NAME;
    } 

    int    size = data.size();
    String _actor = "Actor:";
    String _date = "Date:";
    String _description = "Description:";
    String _separator = "-----------------------------------------------------------------";
    String _attrchanged = "$0 was:$1; $0 now:$2";

    if (seance == null) {
      getLog().warn("Seance is not set, using hard-coded resources");
    } else {
      String[] resKeys = new String[] {
        "functionalities.SWAttributeHistoryFunctionalityLogic.actor", 
        "functionalities.SWAttributeHistoryFunctionalityLogic.date", 
        "functionalities.SWAttributeHistoryFunctionalityLogic.description", 
        "functionalities.SWAttributeHistoryFunctionalityLogic.separator", 
        "functionalities.SWAttributeHistoryFunctionalityLogic.attrchanged"
      };
      String[] res = seance.localizeStrings(resKeys);

      if (!res[0].equals(resKeys[0])) {
        _actor = res[0];
      } 

      if (!res[1].equals(resKeys[1])) {
        _date = res[1];
      } 

      if (!res[2].equals(resKeys[2])) {
        _description = res[2];
      } 

      if (!res[3].equals(resKeys[3])) {
        _separator = res[3];
      } 

      if (!res[4].equals(resKeys[4])) {
        _attrchanged = res[4];
      } 
    } 

    getLog().debug("roDataReading: processing " + size + " records");

    for (int i = 0; i < size; i++) {
      SWRecordPK   recordPK = data.getSWRecordPK(i);
      HashMap      fieldValues = data.getRowAsMap(i);
      StringBuffer history = new StringBuffer();  // record history

      // get all attribute modifications
      Collection   historyRecords = null;

      try {
        historyRecords = attributeHistoryLogRecordHome.findBySWRecordPK(recordPK);
      } catch (Throwable e) {
        getLog().error("Error getting history of the record with pk=" + recordPK, e);

        throw new EJBException2(e, "Error getting history of the record with pk=" + recordPK);
      } 

      boolean isFirst = true;

      for (Iterator it = historyRecords.iterator(); it.hasNext(); ) {
        SWAttributeHistoryLogRecord logRecord = (SWAttributeHistoryLogRecord) it.next();

        if (!isFirst) {
          history.append(lineSep).append(_separator).append(lineSep);
        } else {
          isFirst = false;
        } 

        Long       userID = logRecord.getUserID();

        // todo: time manipulations
        Date       actionDate = new Date(logRecord.getActionDate().longValue());
        SWGroupsPK gpk = new SWGroupsPK(userID.longValue());
        String     actorsName = null;

        try {
          actorsName = usersHome.findByPrimaryKey(new SWUsersPK(gpk)).getName();
        } catch (Throwable e) {
          getLog().error("Error getting actor's name, using user ID instead", e);

          actorsName = userID.toString();
        } 

        String changeDescription = null;

        try {
          changeDescription = getChangeDescription(_attrchanged, logRecord);
        } catch (RemoteException e) {
          getLog().fatal("Communication error", e);

          throw new EJBException2(e);
        } catch (FinderException e) {
          getLog().error("Error getting attribute changes history", e);

          changeDescription = "N/A";
        } 

        history.append(_actor).append(" ").append(actorsName).append(lineSep);
        history.append(_date).append(" ").append(actionDate).append(lineSep);
        history.append(_description).append(lineSep).append(changeDescription);
      } 

      fieldValues.put(columnName, history.toString());
      data.setRowAsMap(i, fieldValues);
    } 

    return data;
  } 

  /**
   * Gets an array of names of tables that must be dropped when functionality is removed.
   * Here we mean not <code>SWTable</code> but simple database tables. In case if functionality
   * wants to kill <code>SWTable</code> it must invoke <code>SWTable.remove()</code> itself
   * (for example it it's own <code>ejbRemove()</code>) and return table's name here.
   * 
   * This method helps to create more or less safe way to remove functionalities. The problem is in
   * transactions handling: mySQL (and most other DBMS) can't rollback DDL commands (create table, drop
   * table, alter table). Thus there can occur a situation when dictionary removes a number of fuctionalities
   * associated with some table before removing the table itself and these functionalities wants to drop
   * other tables. If funtionalities will invoke <code>SWInternalSQL.qlDDL(...)</code> themselves and
   * some functionality will encounter allpication-level exception then the overall transaction
   * cant' be correctly rolled back and this will lead to DB inconsistency.
   * 
   * The solution is to execute all DDL instructions at the end of transaction. In this case all aplliction
   * level exceptions can be safely processed and transaction can be rolled back.
   * 
   * @return Array of names of tables that must be dropped when functionality is removed.
   */
  public String[] getTablesToDropOnRemove() throws RemoteException {
    AHstTableManager aHstTableManager = AHstTableManagerFactory.getAHstTableManager(getSeance());

    return aHstTableManager.getTablesToDropOnRemove(tableID);
  } 

  // _______________________private_______________________
  private String getChangeDescription(String _attrChanged, 
                                      SWAttributeHistoryLogRecord logRecord) throws RemoteException, FinderException {
    StringBuffer res = new StringBuffer();
    Long         recordID = ((SWAttributeHistoryLogRecordPK) logRecord.getPrimaryKey()).getRecordID();
    boolean      isFirst = true;
    Collection   changes = attributeChangeHome.findByRecordID(recordID);

    for (Iterator it = changes.iterator(); it.hasNext(); ) {
      SWAttributeChange change = (SWAttributeChange) it.next();

      if (!isFirst) {
        res.append(lineSep);
      } else {
        isFirst = false;
      } 

      String   attrName = change.getAttributeName();
      String   oldValue = change.getOldValue();
      String   newValue = change.getNewValue();
      Object[] p = new Object[] {
        attrName, oldValue, newValue
      };

      res.append(StringParams.parse(_attrChanged, p));
    } 

    return res.toString();
  } 

  private void getHomes(HomesGetter homesGetter) throws NoSuchFieldException, IllegalAccessException, NamingException {
    if (attributeHistoryLogRecordHome == null) {
      attributeHistoryLogRecordHome = 
        (SWAttributeHistoryLogRecordHome) homesGetter.getHome(SWAttributeHistoryLogRecordHome.class);
    } 

    if (attributeChangeHome == null) {
      attributeChangeHome = (SWAttributeChangeHome) homesGetter.getHome(SWAttributeChangeHome.class);
    } 

    if (usersHome == null) {
      usersHome = (SWUsersHome) homesGetter.getHome(SWUsersHome.class, "swusers/SWUsers");
    } 
  } 

}

