/*
 * SupportWizard 2
 * 
 * Copyright (C) 2001 ISC corp. All Rights Reserved.
 * 
 * $Id: SWBlobFunctionalityLogic.java,v 1.24 2003/08/16 10:23:47 oleg Exp $
 */
package com.supportwizard.functionalities.blob.ejb;

import java.rmi.RemoteException;
import java.sql.ResultSet;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.NameParser;
import javax.naming.Name;
import com.supportwizard.functionalities.SWFunctionalityLogic;
import com.supportwizard.functionalities.SWFunctionalityLogicBase;
import com.supportwizard.functionalities.SWFunctionalityException;
import com.supportwizard.db.Types;
import com.supportwizard.db.ColumnList;
import com.supportwizard.db.TableList;
import com.supportwizard.db.SWInternal_SQL;
import com.supportwizard.db.SW_SQLException;
import com.supportwizard.dictionary.SWDictionaryMeta;
import com.supportwizard.dictionary.SWDictionaryMetaHome;
import com.supportwizard.dictionary.SWDictionaryMetaData;
import com.supportwizard.dictionary.SWTable;
import com.supportwizard.dictionary.SWTableHome;
import com.supportwizard.dictionary.SWColumn;
import com.supportwizard.dictionary.SWDomain;
import com.supportwizard.dictionary.SWType;
import com.supportwizard.dml.interfaces.SWRecord;
import com.supportwizard.dml.SWData;
import com.supportwizard.functionalities.blob.interfaces.*;
import org.apache.log4j.Category;

/**
 * BLOB functionality business logic implementation. Extending <code>SWFunctionalityLogicBase</code>, it contains all essential callbacks for BLOB functionality.<p>
 * The list of callback implemented is:
 * <ul>
 * <li>recordAdding
 * <li>recordModifying
 * <li>recordRemoving
 * 
 * <li>roDataReading
 * <li>rwDataReading
 * <li>columnRemoving
 * </ul>
 * @author Michael Degtyarev
 * @version $Revision: 1.24 $
 * @see SWBlobFunctionality
 */
public class SWBlobFunctionalityLogic extends SWFunctionalityLogicBase implements SWFunctionalityLogic {
  private Long                 tableID;

  /**
   * BLOB functionality business logic implementation. Extending <code>SWFunctionalityLogicBase</code>, it contains all essential callbacks for BLOB functionality.<p>
   * The list of callback implemented is:
   */
  private SWInternal_SQL       sql;
  private SWDictionaryMeta     dictionaryMeta;
  private SWTable              table;
  private SWBlobRecordHome     blobRecordHome;
  private InitialContext       ctx = null;
  private final static boolean RO = true;
  private final static boolean RW = false;

  public SWBlobFunctionalityLogic(Long funcID, Long tableID) throws SWFunctionalityException {
    super(funcID, tableID);

    getLog().debug("ENTRING SWBlobFunctionalityLogic(" + tableID + ")");

    this.tableID = tableID;

    try {
      ctx = new InitialContext();
    } catch (NamingException ex) {
      getLog().error("Failed to obtain initial JNDI context" + ex);

      throw new SWFunctionalityException(ex.getMessage());
    } 

    try {
      SWTableHome tableHome = (SWTableHome) homesGetter.getHome(SWTableHome.class, "dictionary/SWTable");

      table = tableHome.findByPrimaryKey(tableID);

      SWDictionaryMetaHome dictionaryMetaHome = (SWDictionaryMetaHome) homesGetter.getHome(SWDictionaryMetaHome.class, 
              SWDictionaryMetaHome.JNDI_NAME);

      dictionaryMeta = dictionaryMetaHome.create();
      blobRecordHome = (SWBlobRecordHome) homesGetter.getHome(SWBlobRecordHome.class);
    } catch (NamingException ex) {
      getLog().error("Failed to obtain initial JNDI context in HomesGetter", ex);

      throw new SWFunctionalityException(ex.getMessage());
    } catch (FinderException ex) {
      getLog().error("Failed to find remote interface.", ex);

      throw new SWFunctionalityException(ex.getMessage());
    } catch (CreateException ex) {
      getLog().error("Failed to create an instance.", ex);

      throw new SWFunctionalityException(ex.getMessage());
    } catch (RemoteException ex) {
      getLog().error("Remote exception occured while retrieving remote interface.", ex);

      throw new SWFunctionalityException(ex.getMessage());
    } catch (Throwable t) {
      getLog().error("General exception occured.", t);

      throw new SWFunctionalityException(t.getMessage());
    } 

    try {
      sql = new SWInternal_SQL();
    } catch (SW_SQLException ex) {
      getLog().error("Failed to create an instance of SWInternal_SQL", ex);

      throw new SWFunctionalityException(ex.getMessage());
    } 
  }

  /**
   * Called when new record is about to be added. For each reference field, creates a separate record in auxiliary attachments table. Then replaces the value for reference field from JNDI name to the PK value of this attachment in attachments table.
   * 
   * @throws com.supportwizard.functionalities.SWFunctionalityException If something
   * was wrong during event handling.
   * 
   * @param recordValues Mapping of field names to their values for record creation
   * @return Possibly changed new values
   */
  public Map recordAdding(int operationInitiator, Map recordValues) throws SWFunctionalityException, RemoteException {
    getLog().debug("ENTERING recordAdding(" + recordValues + ")");
    super.recordAdding(operationInitiator, recordValues);

    try {
      NameParser           parser = ctx.getNameParser("");
      SWDictionaryMetaData metaData = dictionaryMeta.describeTable(tableID.longValue());
      int                  columnCount = metaData.getColumnCount();

      for (int i = 1; i <= columnCount; i++) {
        String typeName = metaData.getColumnTypeName(i);

        if (typeName.equals(Types.REF) == false) {

          // Not our client :)
          continue;
        } 

        String       columnName = metaData.getColumnName(i);
        Object       columnValue = recordValues.get(columnName);
        String       jndiNameStr = columnValue.toString();
        Name         jndiName = parser.parse(jndiNameStr);
        int          nameSize = jndiName.size();
        String       fileName = jndiName.get(nameSize - 1);
        Object       blobContent = ctx.lookup(jndiName);
        SWBlobRecord blobRecord = blobRecordHome.create(table, fileName, blobContent);

        getLog().debug("Blob record created:" + blobRecord);

        SWBlobRecordPK blobRecordPK = (SWBlobRecordPK) blobRecord.getPrimaryKey();

        getLog().debug("Blob record PK created:" + blobRecordPK);

        Integer blobID = blobRecordPK.getBlobID();

        recordValues.put(columnName, blobID);
        getLog().debug("New values put" + blobRecordPK);
      } 

      getLog().debug("Record values returned:" + recordValues);

      return recordValues;
    } catch (Exception ex) {
      ex.printStackTrace();

      throw new SWFunctionalityException(ex.getMessage());
    } 
  } 

  /**
   * Called when table record is about to be modified.
   * At this point old values can still be obtained using
   * <code>record.getValue()</code> call.
   * 
   * @param record Record that is about to be modified.
   * @param newValue New record value.
   * @throws com.supportwizard.functionalities.SWFunctionalityException If something
   * was wrong during event handling.
   * @return Possibly changed new values.
   */
  public Map recordModifying(int operationInitiator, SWRecord record, 
                             Map newValues) throws SWFunctionalityException, RemoteException {
    getLog().debug("ENTERING recordModifying(" + record + "," + newValues + ")");
    super.recordModifying(operationInitiator, record, newValues);

    try {
      NameParser           parser = ctx.getNameParser("");
      SWDictionaryMetaData metaData = dictionaryMeta.describeTable(tableID.longValue());
      int                  columnCount = metaData.getColumnCount();
      HashMap              oldValues = record.getFieldValues();

      getLog().debug("Old values obtained:" + oldValues);

      for (int i = 1; i <= columnCount; i++) {
        String typeName = metaData.getColumnTypeName(i);

        if (typeName.equals(Types.REF) == false) {

          // Not our client :)
          continue;
        } 

        String columnName = metaData.getColumnName(i);

        if (newValues.containsKey(columnName) == false) {
          continue;
        } 

        Object newColumnValue = newValues.get(columnName);

        if (newColumnValue == null) {
          throw new NullPointerException("Null jndi name passed as a value for the field. Field name:" + columnName);
        } 

        String jndiNameStr = newColumnValue.toString();

        getLog().debug("JNDI name string obtained:" + jndiNameStr);

        Name   jndiName = parser.parse(jndiNameStr);
        int    nameSize = jndiName.size();
        String fileName = jndiName.get(nameSize - 1);
        Object blobContent = ctx.lookup(jndiName);

        getLog().debug(">>> New content:" + blobContent);
        getLog().debug("Old values class:" + oldValues.getClass() + " Old values:" + oldValues + " column name:" 
                       + columnName);

        Integer blobID = (Integer) oldValues.get(columnName);

        getLog().debug("blob ID obtained:" + blobID);

        SWBlobRecordPK blobRecordPK = new SWBlobRecordPK(tableID, blobID);
        SWBlobRecord   blobRecord = blobRecordHome.findByPrimaryKey(blobRecordPK);

        blobRecord.setFileName(fileName);
        blobRecord.setBlobContent(blobContent);
        newValues.put(columnName, blobID);
      } 

      return newValues;
    } catch (Exception ex) {
      throw new SWFunctionalityException(ex.getMessage());
    } 
  } 

  /**
   * Called when the record is about to be removed.
   * 
   * @param record Record that is about to be removed.
   * @throws com.supportwizard.functionalities.SWFunctionalityException If something
   * was wrong during event handling.
   */
  public void recordRemoving(int operationInitiator, SWRecord record) throws SWFunctionalityException, RemoteException {
    getLog().debug("ENTERING recordRemoving(" + record + ")");
    super.recordRemoving(operationInitiator, record);

    try {
      NameParser           parser = ctx.getNameParser("");
      SWDictionaryMetaData metaData = dictionaryMeta.describeTable(tableID.longValue());
      int                  columnCount = metaData.getColumnCount();
      HashMap              fieldValues = record.getFieldValues();

      for (int i = 1; i <= columnCount; i++) {
        String typeName = metaData.getColumnTypeName(i);

        if (typeName.equals(Types.REF) == false) {

          // Not our client :)
          continue;
        } 

        String         columnName = metaData.getColumnName(i);
        Object         columnValue = fieldValues.get(columnName);
        Integer        blobID = (Integer) columnValue;
        SWBlobRecordPK blobRecordPK = new SWBlobRecordPK(tableID, blobID);
        SWBlobRecord   blobRecord = blobRecordHome.findByPrimaryKey(blobRecordPK);

        blobRecord.remove();
      } 
    } catch (Exception ex) {
      throw new SWFunctionalityException(ex.getMessage());
    } 
  } 

  /**
   * Called on reading read-only data. Usually for displaying the whole table
   * in red-only mode. Thus it is unnecessary to fill in any additional SWData's
   * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
   * However, functionality logic may modify some fields or even add virtual fields.
   * 
   * @param data Data being read. All functionalities will in series receive the same SWData instance.
   */
  public SWData roDataReading(SWData data) throws RemoteException, SWFunctionalityException {
    getLog().debug("ENTERING roDataReading(" + data + ")");
    super.roDataReading(data);

    try {
      data = dataReading(data, RO);

      return data;
    } catch (RemoteException ex) {
      getLog().error(ex);

      throw ex;
    } catch (Exception ex) {
      getLog().error(ex);

      throw new RemoteException("Exception caught.", ex);
    } 
  } 

  /**
   * Called on reading some database data that may be modified by user. Usually for displaying
   * and modifying single line from the table. At this point functionality logic can insert some additional
   * info into SWData's
   * {@link com.supportwizard.dml.SWData#setObject(java.lang.String, java.lang.Object) objects}.
   * Functionality logic may aslo modify some fields or add virtual fields.
   * 
   * @param data Data being read. All functionalities will in series receive the same SWData instance.
   */
  public SWData rwDataReading(SWData data) throws RemoteException, SWFunctionalityException {
    getLog().debug("ENTERING rwDataReading(" + data + ")");
    super.rwDataReading(data);

    try {
      data = dataReading(data, RW);

      return data;
    } catch (RemoteException ex) {
      getLog().error(ex);

      throw ex;
    } catch (Exception ex) {
      ex.printStackTrace();
      getLog().error(ex);

      throw new RemoteException("Exception caught.", ex);
    } 
  } 

  public boolean affectsRwData() {
    return true;
  } 

  public void columnRemoving(SWColumn column) throws SWFunctionalityException, RemoteException {
    getLog().debug("ENTERING columnRemoving(" + column + ")");
    super.columnRemoving(column);

    try {
      SWDomain domain = column.getSWDomain();
      SWType   type = domain.getSWType();
      String   typeName = type.getName();

      getLog().debug("Type name fetched is:" + typeName);

      if (typeName.equals(Types.REF) == false) {
        return;
      } 

      /*
       * Prepare to proceed following statement:
       * SELECT blob_column_name FROM user_table
       * For efficiency sake, we aren't going to use here SWRecord stuff.
       */
      String     columnDBName = column.getDBName();
      String     tableDBName = table.getDBName();
      HashMap    params = new HashMap();
      ColumnList columns = new ColumnList();

      columns.add(columnDBName);

      TableList from = new TableList();

      from.add(tableDBName, "");
      params.put("columns", columns);
      params.put("from", from);
      sql.dbConnect();

      try {
        ResultSet rs = sql.sqlQuery(params);

        while (rs.next()) {
          Integer        blobID = (Integer) rs.getObject(1);
          SWBlobRecordPK blobRecordPK = new SWBlobRecordPK(tableID, blobID);
          SWBlobRecord   blobRecord = blobRecordHome.findByPrimaryKey(blobRecordPK);

          blobRecord.remove();
        } 

        rs.close();
      } 
      finally {
        sql.dbDisconnect();
      } 
    } catch (Exception ex) {
      getLog().error(ex);

      throw new SWFunctionalityException(ex.getMessage());
    } 
  } 

  private SWData dataReading(SWData data, boolean readMode) throws Exception {
    SWDictionaryMetaData metaData = dictionaryMeta.describeTable(tableID.longValue());
    int                  columnCount = metaData.getColumnCount();
    HashSet              refColumns = new HashSet();

    for (int i = 1; i <= columnCount; i++) {
      String typeName = metaData.getColumnTypeName(i);

      if (typeName.equals(Types.REF) == false) {

        // Not our client :)
        continue;
      } 

      String columnName;

      if (readMode == RO) {
        columnName = metaData.getDBName(i);
      } else {
        columnName = metaData.getColumnName(i);
      } 

      refColumns.add(columnName);
    } 

    int swDataSize = data.size();

    for (int i = 0; i < swDataSize; i++) {
      HashMap rowData = data.getRowAsMap(i);

      for (Iterator it = refColumns.iterator(); it.hasNext(); ) {
        String  refColumnName = (String) it.next();
        Object  refColumnValue = rowData.get(refColumnName);
        Integer blobID = (Integer) refColumnValue;

        getLog().debug("tableID:" + tableID + " blobID:" + blobID + " refColumnName:" + refColumnName + " rowData:" 
                       + rowData);

        SWBlobRecordPK blobRecordPK = new SWBlobRecordPK(tableID, blobID);
        SWBlobRecord   blobRecord = blobRecordHome.findByPrimaryKey(blobRecordPK);
        String         fileName = blobRecord.getFileName();
        String         fileName_Key = refColumnName + SWBlobFunctionalityBean.BLOB_FILENAME_SUFFIX;

        rowData.put(fileName_Key, fileName);
      } 

      data.setRowAsMap(i, rowData);
    } 

    return data;
  } 

}

